import logging
from typing import Dict, Any

import numpy as np
from seqeval.metrics.sequence_labeling import get_entities

import kolibri
from kolibri.data.corpus.generators import DataGenerator
from kolibri.dnn.embeddings import DefaultEmbedding
from kolibri.dnn.tasks.text.base_model import TextBaseModel
from kolibri.indexers import TextIndexer
from kolibri.logger import logger
from kolibri.metrics.sequence_labeling import sequence_labeling_report


class BaseLabelingModel(TextBaseModel):
    """
    Abstract Labeling Model
    """

    def __init__(self,
                 embedding=None,
                 sequence_length: int = None,
                 hyper_parameters: Dict[str, Dict[str, Any]] = None):
        """

        Args:
            embedding: embedding object
            sequence_length: target sequence length_train
            hyper_parameters: hyper_parameters to overwrite
        """
        super(BaseLabelingModel, self).__init__()
        if embedding is None:
            embedding = DefaultEmbedding()  # type: ignore

        if hyper_parameters is None:
            hyper_parameters = self.get_default_hyper_parameters()

        self.tf_model = None
        self.embedding = embedding
        self.hyper_parameters = hyper_parameters
        self.sequence_length = sequence_length
        self.content_indexer = TextIndexer()
        self.label_indexer = TextIndexer(build_in_vocab='labeling',
                                         min_count=1,
                                         build_vocab_from_labels=True)

    def build_model(self, x_data, y_data):
        """
        Build Model with x_data and y_data

        This function will setup a :class:`CorpusGenerator`,
         then call :meth:`BaseTextClassificationModel.build_model_gen` for preparing processor and model

        Args:
            x_data:
            y_data:

        Returns:

        """

        train_gen = DataGenerator(x_data, y_data)
        self.build_model_generator([train_gen])

    def build_model_generator(self, generators):
        if not self.content_indexer.vocab2idx:
            self.content_indexer.build_vocab_generator(generators)
        self.label_indexer.build_vocab_generator(generators)
        self.embedding.setup_text_processor(self.content_indexer)

        if self.sequence_length is None:
            self.sequence_length = self.embedding.get_seq_length_from_corpus(generators)

        if self.tf_model is None:
            self.build_model_arc()
            self.compile_model()

    def build_model_arc(self) -> None:
        raise NotImplementedError

    def compile_model(self,
                      loss: Any = None,
                      optimizer: Any = None,
                      metrics: Any = None,
                      **kwargs: Any) -> None:
        """
        Configures the model for training.
        call :meth:`tf.keras.Model.predict` to compile model with custom loss, optimizer and metrics
        """
        if loss is None:
            loss = 'sparse_categorical_crossentropy'
        if optimizer is None:
            optimizer = 'adam'
        if metrics is None:
            metrics = ['accuracy']

        self.tf_model.compile(loss=loss,
                              optimizer=optimizer,
                              metrics=metrics,
                              **kwargs)

    def predict(self,
                x_data,
                *,
                batch_size: int = 32,
                truncating: bool = False,
                predict_kwargs: Dict = None):
        """
        Generates output predictions for the input samples.

        Computation is done in batches.

        Args:
            x_data: The input texts, as a Numpy array (or list of Numpy arrays if the model has multiple inputs).
            batch_size: Integer. If unspecified, it will default to 32.
            truncating: remove values from sequences larger than `model.embedding.sequence_length`
            predict_kwargs: arguments passed to :meth:`tf.keras.Model.predict`

        Returns:
            array(s) of predictions.
        """
        if predict_kwargs is None:
            predict_kwargs = {}
        with kolibri.dnn.custom_object_scope():
            if truncating:
                seq_length = self.sequence_length
            else:
                seq_length = None
            tensor = self.content_indexer.transform(x_data,
                                                    seq_length=seq_length,
                                                    max_position=self.embedding.max_position)
            logger.debug('predict seq_length: {}, input: {}'.format(seq_length, np.array(tensor).shape))
            pred = self.tf_model.predict(tensor, batch_size=batch_size, **predict_kwargs)
            pred = pred.argmax(-1)
            lengths = [len(sen) for sen in x_data]

            res = self.label_indexer.inverse_transform(pred,  # type: ignore
                                                       lengths=lengths)
            logger.debug('predict output: {}'.format(np.array(pred).shape))
            logger.debug('predict output argmax: {}'.format(pred.argmax(-1)))
        return res

    def predict_entities(self,
                         x_data,
                         batch_size: int = 32,
                         join_chunk: str = ' ',
                         truncating: bool = False,
                         predict_kwargs: Dict = None):
        """Gets entities from sequence.

        Args:
            x_data: The input texts, as a Numpy array (or list of Numpy arrays if the model has multiple inputs).
            batch_size: Integer. If unspecified, it will default to 32.
            truncating: remove values from sequences larger than `model.embedding.sequence_length`
            join_chunk: str or False,
            predict_kwargs: arguments passed to :meth:`tf.keras.Model.predict`

        Returns:
            list: list of entity.
        """
        if isinstance(x_data, tuple):
            text_seq = x_data[0]
        else:
            text_seq = x_data
        res = self.predict(x_data,
                           batch_size=batch_size,
                           truncating=truncating,
                           predict_kwargs=predict_kwargs)
        new_res = [get_entities(seq) for seq in res]
        final_res = []
        for index, seq in enumerate(new_res):
            seq_data = []
            for entity in seq:
                res_entities = []
                for i, e in enumerate(text_seq[index][entity[1]:entity[2] + 1]):
                    # Handle bert tokenizer
                    if e.startswith('##') and len(res_entities) > 0:
                        res_entities[-1] += e.replace('##', '')
                    else:
                        res_entities.append(e)
                value = None
                if join_chunk is False:
                    value = res_entities
                else:
                    value = join_chunk.join(res_entities)

                seq_data.append({
                    "entity": entity[0],
                    "start": entity[1],
                    "end": entity[2],
                    "value": value,
                })

            final_res.append({
                'tokenized': x_data[index],
                'y_values': seq_data
            })
        return final_res

    def evaluate(self,
                 x_data, y_data,
                 batch_size: int = 32,
                 digits: int = 4,
                 truncating: bool = False) -> Dict:
        """
        Build a text report showing the main labeling metrics.
        """
        y_pred = self.predict(x_data,
                              batch_size=batch_size,
                              truncating=truncating)
        y_true = [seq[:len(y_pred[index])] for index, seq in enumerate(y_data)]

        new_y_pred = []
        for x in y_pred:
            new_y_pred.append([str(i) for i in x])
        new_y_true = []
        for x in y_true:
            new_y_true.append([str(i) for i in x])

        report = sequence_labeling_report(y_true, y_pred, digits=digits)
        return report


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    from kolibri.data.corpus.corpus import CONLL2003ENCorpus

    corpus = CONLL2003ENCorpus('train')
    train_x, train_y = corpus.get_data()
    valid_x, valid_y = CONLL2003ENCorpus('valid')

    train_x, train_y = train_x[:5120], train_y[:5120]

    model = None
    model.build_model(train_x[:100], train_y[:100])

    model.fit(train_x[:1000], train_y[:1000], epochs=10)

    model.evaluate(train_x[:20], train_y[:20])
    print("Hello world")



import copy

import numpy as np
from scipy.sparse import hstack
from sklearn.linear_model import LogisticRegression

from kolibri.indexers.label_indexer import LabelIndexer
from kolibri.kolibri_component import Component
from kolibri.utils.misc import check_random_state


class ChainEstimator(Component):
    """ Classifier Chains for multi-label learning.

    Parameters
    ----------
    base_estimator: skmultiflow.core.BaseSKMObject or sklearn.BaseEstimator (default=LogisticRegression)
        Each member of the ensemble is an instance of the base estimator

    order : str (default=None)
        `None` to use default order, 'random' for random order.

    random_state: int, RandomState instance or None, optional (default=None)
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used by `np.random`.




    Notes
    -----
    Classifier Chains [1]_ is a popular method for multi-label learning. It exploits correlation between y_values
    by incrementally building binary classifiers for each label.

    scikit-learn also includes 'ClassifierChain'. A difference is probabilistic extensions are included here.


    References
    ----------
    .. [1] Read, Jesse, Bernhard Pfahringer, Geoff Holmes, and Eibe Frank. "Classifier chains for multi-label
       classification." In Joint European Conference on Machine Learning and Knowledge Discovery in Databases,
       pp. 254-269. Springer, Berlin, Heidelberg, 2009.

    """

    # TODO: much of this can be shared with Regressor Chains, probably should use a base class to inherit here.

    def __init__(self, base_estimator=LogisticRegression(), order=None, indexer=None, random_state=None):
        super().__init__()
        self.base_estimator = base_estimator
        self.order = order

        self.chain = None
        self.ensemble = None
        self.L = None
        self.random_state = random_state  # This is the actual random_state object used internally
        if indexer:
            self.indexer = indexer
        else:
            self.indexer = LabelIndexer(multi_label=True)

        self.__configure()

    def __configure(self):
        self.ensemble = None
        self.L = -1
        self.random_state = check_random_state(self.random_state)

    def fit(self, X, y, classes=None, sample_weight=None):
        """ Fit the model.

        Parameters
        ----------
        X : numpy.ndarray of shape (n_samples, n_features)
            The features to train the model.

        y: numpy.ndarray of shape (n_samples, n_targets)
            An array-like with the y_values of all samples in X.

        classes: Not used (default=None)

        sample_weight: Not used (default=None)

        Returns
        -------
        self

        """

        self.indexer.build_label_dict([[cl] for cl in classes])
        y = np.array(self.indexer.numerize_label_sequences(y))

        N, self.L = y.shape
        L = self.L
        N, D = X.shape

        self.chain = np.arange(L)
        if self.order == 'random':
            self.random_state.shuffle(self.chain)

        # Set the chain order
        y = y[:, self.chain]

        # Train
        self.ensemble = [copy.deepcopy(self.base_estimator) for _ in range(L)]
        for j in range(self.L):
            if j > 0:
                XY = hstack([XY, np.array([y[:, j - 1]]).T])
            else:
                XY = X
            self.ensemble[j].fit(XY, y[:, j])
        return self

    def partial_fit(self, X, y, classes=None, sample_weight=None):
        """ Partially (incrementally) fit the model.

        Parameters
        ----------
        X : numpy.ndarray of shape (n_samples, n_features)
            The features to train the model.

        y: numpy.ndarray of shape (n_samples)
            An array-like with the y_values of all samples in X.

        classes: Not used (default=None)

        sample_weight: NOT used (default=None)

        Returns
        -------
        self

        """
        if self.ensemble is None:
            # This is the first time that the model is fit
            self.fit(X, y, classes=classes)
            return self

        y = np.array(self.indexer.numerize_label_sequences(y))

        N, self.L = y.shape

        # Set the chain order
        y = y[:, self.chain]

        for j in range(self.L):
            if j > 0:
                XY = hstack([XY, np.array([y[:, j - 1]]).T])
            else:
                XY = X
            self.ensemble[j].partial_fit(XY, y[:, j])
        return self

    def predict(self, X, return_labels=True):
        """ Predict classes for the passed texts.

        Parameters
        ----------
        X : numpy.ndarray of shape (n_samples, n_features)
            The set of texts samples to predict the y_values for.

        Returns
        -------
        A numpy.ndarray with all the predictions for the samples in X.

        """
        N, D = X.shape
        Y = np.zeros((N, self.L))

        for j in range(self.L):
            if j > 0:
                X = hstack([X, Y[:, j - 1]])
            Y[:, j] = self.ensemble[j].predict(X)

        # Unset the chain order (back to default)
        predictions = Y[:, np.argsort(self.chain)]

        if return_labels:
            return self.indexer.inverse_transform(predictions)
        return predictions

    def predict_proba(self, X):
        """ Estimates the probability of each sample in X belonging to each of the class-y_values.

        Parameters
        ----------
        X : numpy.ndarray of shape (n_samples, n_features)
            The matrix of samples one wants to predict the class probabilities for.

        Returns
        -------
        A numpy.ndarray of shape (n_samples, n_labels), in which each outer entry is associated with the X entry of the
        same index. And where the list in index [i] contains len(self.target_values) elements, each of which represents
        the probability that the i-th sample of X belongs to a certain class-label.

        Notes
        -----
        Returns marginals [P(y_1=1|x),...,P(y_L=1|x,y_1,...,y_{L-1})]
        i.e., confidence predictions given inputs, for each instance.

        This function suitable for multi-label (binary) texts
        only at the moment (may give index-out-of-bounds error if
        uni- or multi-label (of > 2 values) texts is used in training).
        """
        N, D = X.shape
        Y = np.zeros((N, self.L))
        for j in range(self.L):
            if j > 0:
                X = np.column_stack([X, Y[:, j - 1]])
            Y[:, j] = self.ensemble[j].predict_proba(X)[:, 1]
        return Y

    def reset(self):
        self.__configure()
        return self

    def _more_tags(self):
        return {'multioutput': True,
                'multioutput_only': True}


def P(y, x, cc, payoff=np.prod):
    """ Payoff function, P(Y=y|X=x)

        What payoff do we get for predicting y | x, under model cc.

        Parameters
        ----------
        x: input instance
        y: its true y_values
        cc: a classifier chain
        payoff: payoff function

        Returns
        -------
        A single number; the payoff of predicting y | x.
    """
    D = x.shape[1]
    L = len(y)

    p = np.zeros(L)
    #    for j in range(L):
    #        P_j = cc.ensemble[j].predict_proba(xy[0:D+j].reshape(1, -1))[0]  # e.g., [0.9, 0.1] wrt 0, 1
    #        xy[D+j] = y[j]                                           # e.g., 1
    #        p[j] = P_j[y[j]]                                         # e.g., 0.1

    for j in range(L):
        if j > 0:
            xy = hstack([xy, y[j - 1]])
        else:
            xy = x

        P_j = cc.ensemble[j].predict_proba(xy)[0]
        # e.g., 1
        p[j] = P_j[y[j]]  # e.g., 0.1
        # or, y[j] = 0 is predicted with probability p[j] = 0.9
    return payoff(p)


class ProbabilisticChainEstimator(ChainEstimator):
    """ Probabilistic Classifier Chains for multi-label learning.

    Published as 'PCC'

    Parameters
    ----------
    base_estimator: skmultiflow or sklearn model (default=LogisticRegression)
        This is the ensemble classifier type, each ensemble classifier is going
        to be a copy of the base_estimator.

    order : str (default=None)
        `None` to use default order, 'random' for random order.

    random_state: int, RandomState instance or None, optional (default=None)
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used by `np.random`.


    """

    def __init__(self, base_estimator=LogisticRegression(), order=None, indexer=None):
        super().__init__(base_estimator=base_estimator, order=order, indexer=indexer)

    def predict(self, X, return_labels=True):
        """ Predict classes for the passed texts.

        Parameters
        ----------
        X : numpy.ndarray of shape (n_samples, n_features)
            The set of texts samples to predict the y_values for.

        Returns
        -------
        A numpy.ndarray with all the predictions for the samples in X.

        Notes
        -----
        Explores all possible branches of the probability tree (i.e., all possible 2^L label combinations).
        """

        N, D = X.shape
        Yp = np.zeros((N, self.L))

        # for each instance
        for n in range(N):
            w_max = 0.
            # for each and every possible label combination
            for b in range(2 ** self.L):
                # put together a label vector
                y_ = np.array(list(map(int, np.binary_repr(b, width=self.L))))
                # ... and gauge a probability for it (given x)
                w_ = P(y_, X[n], self)
                # if it performs well, keep it, and record the max
                if w_ > w_max:
                    Yp[n, :] = y_[:].copy()
                    w_max = w_

        if return_labels:
            return self.indexer.reverse_numerize_label_sequences(Yp)
        return Yp


class MonteCarloChainEstimator(ProbabilisticChainEstimator):
    """ Monte Carlo Sampling Classifier Chains for multi-label learning.

        PCC, using Monte Carlo sampling, published as 'MCC'.

        M samples are taken from the posterior distribution. Therefore we need
        a probabilistic interpretation of the output, and thus, this is a
        particular variety of ProbabilisticClassifierChain.

    Parameters
    ----------
    base_estimator: StreamModel or sklearn model
        This is the ensemble classifier type, each ensemble classifier is going
        to be a copy of the base_estimator.

    M: int (default=10)
        Number of samples to take from the posterior distribution.

    random_state: int, RandomState instance or None, optional (default=None)
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used by `np.random`.
    """

    def __init__(self, base_estimator=LogisticRegression(), M=10, indexer=None):
        # Do M iterations, unless overridden by M at prediction time
        ChainEstimator.__init__(self, base_estimator, indexer=indexer)
        self.M = M

    def sample(self, x):
        """
            Sample y ~ P(y|x)

            Returns
            -------
            y: a sampled label vector
            p: the associated probabilities, i.e., p(y_j=1)=p_j
        """
        D = x.shape[1]

        p = np.zeros(self.L)
        y = np.zeros(self.L)

        for j in range(self.L):
            if j > 0:
                xy = hstack([xy, y[j - 1]])
            else:
                xy = x

            P_j = self.ensemble[j].predict_proba(xy)[0]
            y_j = self.random_state.choice(2, 1, p=P_j)
            y[j] = y_j
            p[j] = P_j[y_j]

        return y, p

    def predict(self, X, return_labels=True, M=None):
        """ Predict classes for the passed texts.

        Parameters
        ----------
        X : numpy.ndarray of shape (n_samples, n_features)
            The set of texts samples to predict the y_values for.

        M: int (optional, default=None)
            Number of sampling iterations. If None, M is set equal to the M value used for initialization

        Returns
        -------
        A numpy.ndarray with all the predictions for the samples in X.

        Notes
        -----
        Quite similar to the `ProbabilisticClassifierChain.predict()` function.

        Depending on the implementation, `y_max`, `w_max` may be initially set to 0,
        if we wish to rely solely on the sampling. Setting the `w_max` based on a naive CC prediction
        gives a good baseline to work from.

        """
        N, D = X.shape
        Yp = np.zeros((N, self.L))

        if M is None:
            M = self.M

        # for each instance
        for n in range(N):
            y = ChainEstimator.predict(self, X.getrow(n), False)
            Yp[n, :] = y
            w_max = P(Yp[n, :].astype(int), X.getrow(n), self)
            # for M times
            for m in range(M):
                y_, p_ = self.sample(X.getrow(n))  # N.B. in fact, the calculation p_ is done again in P.
                w_ = P(y_.astype(int), X.getrow(n), self)
                # if it performs well, keep it, and record the max
                if w_ > w_max:
                    Yp[n, :] = y_[:].copy()
                    w_max = w_
        if return_labels:
            return self.indexer.inverse_transform(Yp)
        return Yp



import numbers

import numpy as np


def import_module(dotted_path):
    """
    Imports the specified module based on the
    dot notated import path for the module.
    """
    import importlib

    module_parts = dotted_path.split('.')
    module_path = '.'.join(module_parts[:-1])
    module = importlib.import_module(module_path)

    return getattr(module, module_parts[-1])

def initialize_class(data, *args, **kwargs):
    """
    :param data: A string or dictionary containing a import_path attribute.
    """
    if isinstance(data, dict):
        import_path = data.get('import_path')
        data.update(kwargs)
        Class = import_module(import_path)

        return Class(*args, **data)
    else:
        Class = import_module(data)

        return Class(*args, **kwargs)


def ordered(obj):
    if isinstance(obj, dict):
        return sorted((k, ordered(v)) for k, v in obj.items())
    if isinstance(obj, list):
        return sorted(ordered(x) for x in obj)
    else:
        return obj


def module_path_from_object(o):
    """Returns the fully qualified class path of the instantiated object."""
    return o.__class__.__module__ + "." + o.__class__.__name__


def check_random_state(seed):
    """Turn seed into a np.random.RandomState instance.

    Parameters
    ----------
    seed : None | int | instance of RandomState
        If seed is None, return the RandomState singleton used by np.random.
        If seed is an int, return a new RandomState instance seeded with seed.
        If seed is already a RandomState instance, return it.
        Otherwise raise ValueError.

    Notes
    -----
    Code from sklearn

    """
    if seed is None or seed is np.random:
        return np.random.mtrand._rand
    if isinstance(seed, (numbers.Integral, np.integer)):
        return np.random.RandomState(seed)
    if isinstance(seed, np.random.RandomState):
        return seed
    raise ValueError('{} cannot be used to seed a numpy.random.RandomState instance'.format(seed))


def class_label_statistics(y):
    unique, counts = np.unique(y, return_counts=True)
    class_stats = dict(zip(unique, counts))

    return sorted(class_stats.items(), key=lambda x: -x[1])



from abc import ABCMeta, abstractmethod


class BaseStream(metaclass=ABCMeta):
    """ The abstract class setting up the minimum requirements of a stream.

    Raises
    ------
    NotImplementedError: This is an abstract class.

    """

    def __init__(self):
        self.n_samples = 0
        self.n_targets = 0
        self.n_features = 0
        self.n_classes = 0
        self.current_sample_x = None
        self.current_sample_y = None
        self.content_column = None
        self.target_columns = None
        self.target_values = list()
        self.name = None
        self.random_state = None
        self.task_type = None

    @abstractmethod
    def prepare(self):
        """ prepare_for_use
        Prepare the stream for use. Can be the reading of a file, or
        the generation of a function.
        """
        raise NotImplementedError

    @abstractmethod
    def next(self, batch_size=1):
        """ Generates or returns next `batch_size` samples in the stream.

        """
        raise NotImplementedError

    def last(self):
        """ Retrieves last `batch_size` samples in the stream.
        Returns
        -------
        tuple or tuple list
            A numpy.ndarray of shape (batch_size, n_features) and an array-like of shape
            (batch_size, n_targets), representing the next batch_size samples.
        """
        return self.current_sample_x, self.current_sample_y

    def restart(self):
        """  Restart the stream. """
        self.prepare()

    def get_data(self, nb_samples=None):
        raise NotImplementedError

    def get_next_document(self):
        raise NotImplementedError

    def get_all_documents(self):
        raise NotImplementedError

    def has_more_samples(self):
        """
        Checks if stream has more samples.
        Returns
        -------
        Boolean
            True if stream has more samples.
        """
        return True

    def get_class_type(self):
        return 'stream'



from kolibri.document import Document


class MemoryLearner():
    """
    Allows a chat bot to be trained using a list of strings
    where the list represents a conversation.
    """

    def __init__(self, application):
        self.application = application

    def fit(self, X, y):
        raise NotImplementedError

    def train(self, texts, target_texts, **kwargs):
        """
        Train model based on the provided list of
        documents.
        """

        documents_to_create = []

        for conversation_count, text in enumerate(texts):
            document_search_text = self.application.datastore.tagger.get_text_index_string(text)
            target_document_text = target_texts[conversation_count]
            target_document_search_text = self.application.datastore.tagger.get_text_index_string(target_document_text)

            document = Document(
                text=text,
                search_text=document_search_text,
                target_text=target_document_text,
                search_target_text=target_document_search_text,
                label='training'
            )

            documents_to_create.append(document)

        self.application.datastore.create_many(documents_to_create)



import time
import warnings

import matplotlib.pyplot as plt
from matplotlib.rcsetup import cycler

from kolibri.config import TaskType
from kolibri.utils import constants
from kolibri.utils.data_structures import Deque


class PerformanceVisualizer():
    """ This class is responsible for maintaining and updating the plot modules
    for the evaluators.
    
    It uses `matplotlib.pyplot` modules to create the main plot, which
    depending on the options passed to it as parameter, will create multiple 
    subplots to better display all requested metrics.
    """

    def __init__(self, task_type, n_wait, dataset_name, metrics, n_models, model_names, data_dict):
        super().__init__()

        # Default values
        self._sample_ids = []
        self._is_legend_set = False
        self._frame_cnt = 0
        self._plot_trackers = {}
        self._text_annotations = []
        self._last_draw_timestamp = 0

        # Configuration
        self._data_dict = data_dict
        self.n_wait = n_wait
        self.dataset_name = dataset_name
        self.n_models = n_models

        # Validate inputs
        if task_type is None or task_type == constants.UNDEFINED:
            raise ValueError('Task type for visualizer object is undefined.')
        else:
            if task_type in [TaskType.CLASSIFICATION, TaskType.REGRESSION, TaskType.MULTI_TARGET_CLASSIFICATION,
                             TaskType.MULTI_TARGET_REGRESSION]:
                self.task_type = task_type
            else:
                raise ValueError('Invalid task type: {}'.format(task_type))

        if model_names is None:
            self.model_names = ['M{}'.format(i) for i in range(n_models)]
        else:
            if isinstance(model_names, list):
                if len(model_names) != n_models:
                    raise ValueError("Number of model names {} does not match the number of models {}.".
                                     format(len(model_names), n_models))
                else:
                    self.model_names = model_names
            else:
                raise ValueError("model_names must be a list.")

        if metrics is not None:
            if len(metrics) < 1:
                raise ValueError('The metrics list is empty.')
            else:
                self.metrics = metrics
        else:
            raise ValueError('Invalid metrics {}'.format(metrics))

        if constants.DATA_POINTS in metrics and n_models > 1:
            raise ValueError("Can not use multiple models with texts points visualization.")

        # Proceed with configuration
        self.__configure()

    def on_new_train_step(self, sample_id, data_buffer):
        """ Whenever the EvaluationVisualiser should
        be aware of some new texts, the caller will invoke this function,
        passing the new texts buffer.
        
        Parameters
        ----------
        sample_id: int
            The current sample id.

        data_buffer: EvaluationDataBuffer
            A buffer containing evaluation texts for a single training / visualization step.
            
        Raises
        ------
        ValueError: If an exception is raised during the draw operation.
         
        """

        try:
            current_time = time.time()
            self._clear_annotations()
            self._update_plots(sample_id, data_buffer)

            # To mitigate re-drawing overhead for fast models use frame counter (default = 5 frames).
            # To avoid slow refresh rate in slow models use a time limit (default = 1 sec).
            if (self._frame_cnt < 5) or (current_time - self._last_draw_timestamp > 1):
                plt.subplots_adjust(right=0.72)  # Adjust subplots to include metrics annotations
                self.fig.canvas.draw()
                plt.pause(1e-9)
                self._frame_cnt = 0
                self._last_draw_timestamp = current_time
            else:
                self._frame_cnt += 1
        except BaseException as exception:
            raise ValueError('Failed when trying to draw plot. Exception: {} | Type: {}'.
                             format(exception, type(exception).__name__))

    def __configure(self):
        """  This function will verify which subplots should be create. Initializing
        all relevant objects to keep track of the plotting points.

        """
        font_size_small = 8
        font_size_medium = 10
        font_size_large = 12

        plt.rc('font', size=font_size_small)  # controls default text sizes
        plt.rc('axes', titlesize=font_size_medium)  # font size of the axes title
        plt.rc('axes', labelsize=font_size_small)  # font size of the x and y y_values
        plt.rc('xtick', labelsize=font_size_small)  # font size of the tick y_values
        plt.rc('ytick', labelsize=font_size_small)  # font size of the tick y_values
        plt.rc('legend', fontsize=font_size_small)  # legend font size
        plt.rc('figure', titlesize=font_size_large)  # font size of the figure title

        warnings.filterwarnings("ignore", ".*GUI is implemented.*")
        warnings.filterwarnings("ignore", ".*left==right.*")
        warnings.filterwarnings("ignore", ".*Passing 1d.*")

        self._sample_ids = []

        plt.ion()
        self.fig = plt.figure(figsize=(9, 5))
        self.fig.suptitle(self.dataset_name)
        base = 11 + len(self.metrics) * 100  # 3-digit integer describing the position of the subplot.
        self.fig.canvas.set_window_title('kolibri')

        # Subplots handler
        for metric_id in self.metrics:
            data_ids = self._data_dict[metric_id]
            self._plot_trackers[metric_id] = PlotDataTracker(data_ids)
            plot_tracker = self._plot_trackers[metric_id]
            plot_tracker.sub_plot_obj = self.fig.add_subplot(base)
            base += 1
            if metric_id == constants.TRUE_VS_PREDICTED:
                handle = []
                plot_tracker.sub_plot_obj.set_prop_cycle(cycler('color', ['c', 'm', 'y', 'k']))
                for data_id in data_ids:
                    if data_id == constants.Y_TRUE:
                        # True texts
                        plot_tracker.data[data_id] = []
                        label = 'True value'
                        line_style = '--'
                        line_obj = plot_tracker.line_objs
                        if self.task_type == TaskType.CLASSIFICATION:
                            line_obj[data_id], = plot_tracker.sub_plot_obj.step(self._sample_ids,
                                                                                plot_tracker.data[data_id],
                                                                                label=label, linestyle=line_style)
                        else:
                            line_obj[data_id], = plot_tracker.sub_plot_obj.plot(self._sample_ids,
                                                                                plot_tracker.data[data_id],
                                                                                label=label, linestyle=line_style)
                        handle.append(line_obj[data_id])
                    else:
                        # Predicted texts
                        plot_tracker.data[data_id] = [[] for _ in range(self.n_models)]
                        plot_tracker.line_objs[data_id] = [None for _ in range(self.n_models)]
                        line_obj = plot_tracker.line_objs[data_id]
                        for i in range(self.n_models):
                            label = 'Predicted {}'.format(self.model_names[i])
                            line_style = '--'
                            if self.task_type == TaskType.CLASSIFICATION:
                                line_obj[i], = plot_tracker.sub_plot_obj.step(self._sample_ids,
                                                                              plot_tracker.data[data_id][i],
                                                                              label=label, linestyle=line_style)
                            else:
                                line_obj[i], = plot_tracker.sub_plot_obj.plot(self._sample_ids,
                                                                              plot_tracker.data[data_id][i],
                                                                              label=label, linestyle=line_style)
                            handle.append(line_obj[i])
                plot_tracker.sub_plot_obj.legend(handles=handle, loc=2, bbox_to_anchor=(1.01, 1.))
                plot_tracker.sub_plot_obj.set_title('True vs Predicted')
                plot_tracker.sub_plot_obj.set_ylabel('y')

            elif metric_id == constants.DATA_POINTS:
                plot_tracker.data['buffer_size'] = 100
                plot_tracker.data['X'] = Deque(plot_tracker.data['buffer_size'])
                plot_tracker.data['target_values'] = None
                plot_tracker.data['predictions'] = Deque(plot_tracker.data['buffer_size'])
                plot_tracker.data['clusters'] = []
                plot_tracker.data['clusters_initialized'] = False

            else:
                # Default case, 'mean' and 'current' performance
                handle = []
                sorted_data_ids = data_ids.copy()
                sorted_data_ids.sort()  # For better usage of the color cycle, start with 'current' texts
                for data_id in sorted_data_ids:
                    plot_tracker.data[data_id] = [[] for _ in range(self.n_models)]
                    plot_tracker.line_objs[data_id] = [None for _ in range(self.n_models)]
                    line_obj = plot_tracker.line_objs[data_id]
                    for i in range(self.n_models):
                        if data_id == constants.CURRENT:
                            label = '{}  (current, {} samples)'.format(self.model_names[i], self.n_wait)
                            line_style = '-'
                        else:
                            label = '{} (mean)'.format(self.model_names[i])
                            line_style = ':'
                        line_obj[i], = plot_tracker.sub_plot_obj.plot(self._sample_ids,
                                                                      plot_tracker.data[data_id][i],
                                                                      label=label,
                                                                      linestyle=line_style)
                        handle.append(line_obj[i])
                self._set_fig_legend(handle)

                if metric_id == constants.ACCURACY:
                    plot_tracker.sub_plot_obj.set_title('Accuracy')
                    plot_tracker.sub_plot_obj.set_ylabel('acc')
                elif metric_id == constants.HAMMING_SCORE:
                    plot_tracker.sub_plot_obj.set_title('Hamming score')
                    plot_tracker.sub_plot_obj.set_ylabel('hamming score')
                elif metric_id == constants.HAMMING_LOSS:
                    plot_tracker.sub_plot_obj.set_title('Hamming loss')
                    plot_tracker.sub_plot_obj.set_ylabel('hamming loss')
                elif metric_id == constants.EXACT_MATCH:
                    plot_tracker.sub_plot_obj.set_title('Exact Match')
                    plot_tracker.sub_plot_obj.set_ylabel('exact match')
                elif metric_id == constants.J_INDEX:
                    plot_tracker.sub_plot_obj.set_title('Jaccard Index')
                    plot_tracker.sub_plot_obj.set_ylabel('j-index')
                elif metric_id == constants.MSE:
                    plot_tracker.sub_plot_obj.set_title('Mean Squared Error')
                    plot_tracker.sub_plot_obj.set_ylabel('mse')
                elif metric_id == constants.MAE:
                    plot_tracker.sub_plot_obj.set_title('Mean Absolute Error')
                    plot_tracker.sub_plot_obj.set_ylabel('mae')
                elif metric_id == constants.AMSE:
                    plot_tracker.sub_plot_obj.set_title('Average Mean Squared Error')
                    plot_tracker.sub_plot_obj.set_ylabel('amse')
                elif metric_id == constants.AMAE:
                    plot_tracker.sub_plot_obj.set_title('Average Mean Absolute Error')
                    plot_tracker.sub_plot_obj.set_ylabel('amae')
                elif metric_id == constants.ARMSE:
                    plot_tracker.sub_plot_obj.set_title('Average Root Mean Squared Error')
                    plot_tracker.sub_plot_obj.set_ylabel('armse')
                elif metric_id == constants.DATA_POINTS:
                    plot_tracker.sub_plot_obj.set_title('')
                    plot_tracker.sub_plot_obj.set_xlabel('Feature x')
                    plot_tracker.sub_plot_obj.set_ylabel('Feature y')
                else:
                    plot_tracker.sub_plot_obj.set_title('Unknown metric')
                    plot_tracker.sub_plot_obj.set_ylabel('')

        if constants.DATA_POINTS not in self.metrics:
            plt.xlabel('Samples')

        self.fig.subplots_adjust(hspace=.5)
        self.fig.tight_layout(rect=[0, .04, 1, 0.98], pad=2.6, w_pad=0.5, h_pad=1.0)

    def _set_fig_legend(self, handles=None):
        if not self._is_legend_set:
            self.fig.legend(handles=handles, ncol=self.n_models, bbox_to_anchor=(0.02, 0.0), loc="lower left")
            self._is_legend_set = True

    def _update_plots(self, sample_id, data_buffer):
        self._sample_ids.append(sample_id)
        for metric_id, data_ids in data_buffer.data_dict.items():
            update_xy_limits = True
            y_min = 0.0
            y_max = 1.0
            pad = 0.1  # Default padding to set above and bellow plots
            plot_tracker = self._plot_trackers[metric_id]
            if metric_id == constants.TRUE_VS_PREDICTED:
                # Process true values
                data_id = constants.Y_TRUE
                plot_tracker.data[data_id].append(data_buffer.get_data(metric_id=metric_id, data_id=data_id))
                plot_tracker.line_objs[data_id].set_data(self._sample_ids, plot_tracker.data[data_id])
                # Process predicted values
                data_id = constants.Y_PRED
                data = data_buffer.get_data(metric_id=metric_id, data_id=data_id)
                for i in range(self.n_models):
                    plot_tracker.data[data_id][i].append(data[i])
                    plot_tracker.line_objs[data_id][i].set_data(self._sample_ids, plot_tracker.data[data_id][i])
                    y_min = min([plot_tracker.data[data_id][i][-1], plot_tracker.data[constants.Y_TRUE][-1], y_min])
                    y_max = max([plot_tracker.data[data_id][i][-1], plot_tracker.data[constants.Y_TRUE][-1], y_max])
            elif metric_id == constants.DATA_POINTS:
                update_xy_limits = False
                # Process features
                data_id = 'X'
                features_dict = data_buffer.get_data(metric_id=metric_id, data_id=data_id)
                feature_indices = list(features_dict.keys())
                feature_indices.sort()
                # Store tuple of feature values into the buffer, sorted by index
                plot_tracker.data[data_id].add_element([[features_dict[feature_indices[0]],
                                                         features_dict[feature_indices[1]]]])

                plot_tracker.sub_plot_obj.set_xlabel('Feature {}'.format(feature_indices[0]))
                plot_tracker.sub_plot_obj.set_ylabel('Feature {}'.format(feature_indices[1]))
                # TODO consider a fading/update strategy instead
                plot_tracker.sub_plot_obj.clear()
                X1 = plot_tracker.data[data_id].get_queue()[-1][0]
                X2 = plot_tracker.data[data_id].get_queue()[-1][1]

                # Process label values
                data_id = 'target_values'
                plot_tracker.data[data_id] = data_buffer.get_data(metric_id=metric_id, data_id=data_id)
                if not plot_tracker.data['clusters_initialized']:
                    for j in range(len(plot_tracker.data[data_id])):
                        plot_tracker.data['clusters'].append(Deque(plot_tracker.data['buffer_size']))

                # Process predictions
                data_id = 'predictions'
                plot_tracker.data[data_id].add_element([data_buffer.get_data(metric_id=metric_id, data_id=data_id)])

                for k, cluster in enumerate(plot_tracker.data['clusters']):
                    if plot_tracker.data[data_id].get_queue()[-1] == k:
                        plot_tracker.data['clusters'][k].add_element([(X1, X2)])
                        # TODO confirm buffer update inside the loop
                    if cluster.get_queue():
                        temp = cluster.get_queue()
                        plot_tracker.sub_plot_obj.scatter(*zip(*temp), label="Class {k}".format(k=k))
                plot_tracker.sub_plot_obj.legend(loc=2, bbox_to_anchor=(1.01, 1.))

            else:
                # Default case, 'mean' and 'current' performance
                for data_id in data_ids:
                    # Buffer texts
                    data = data_buffer.get_data(metric_id=metric_id, data_id=data_id)
                    for i in range(self.n_models):
                        plot_tracker.data[data_id][i].append(data[i])
                        plot_tracker.line_objs[data_id][i].set_data(self._sample_ids, plot_tracker.data[data_id][i])
                # Process texts
                for i in range(self.n_models):
                    # Update annotations
                    self._update_annotations(i, plot_tracker.sub_plot_obj, self.model_names[i],
                                             plot_tracker.data[constants.MEAN][i][-1],
                                             plot_tracker.data[constants.CURRENT][i][-1])
                    # Update plot limits
                    if metric_id in [constants.MSE, constants.MAE, constants.AMSE, constants.AMAE, constants.ARMSE]:
                        y_min = -1
                        y_max = max([plot_tracker.data[constants.MEAN][i][-1],
                                     plot_tracker.data[constants.CURRENT][i][-1], y_max])
                        pad = 0.5 * y_max  # Padding bellow and above thresholds
            if update_xy_limits:
                plot_tracker.sub_plot_obj.set_ylim((y_min - pad, y_max + pad))
                plot_tracker.sub_plot_obj.set_xlim(0, self._sample_ids[-1])

    def _clear_annotations(self):
        """ Clear annotations, so next frame is correctly rendered. """
        for i in range(len(self._text_annotations)):
            self._text_annotations[i].remove()
        self._text_annotations = []

    def _update_annotations(self, idx, subplot, model_name, mean_value, current_value):
        xy_pos_default = (1.02, .90)  # Default xy position for metric annotations
        shift_y = 10 * (idx + 1)  # y axis shift for plot annotations
        xy_pos = xy_pos_default
        if idx == 0:
            self._text_annotations.append(subplot.annotate('{: <12} | {: ^16} | {: ^16}'.
                                                           format('Model', 'Mean', 'Current'),
                                                           xy=xy_pos, xycoords='axes fraction'))
        self._text_annotations.append(subplot.annotate('{: <10.10s}'.format(model_name[:6]),
                                                       xy=xy_pos, xycoords='axes fraction',
                                                       xytext=(0, -shift_y), textcoords='offset points'))
        self._text_annotations.append(subplot.annotate('{: ^16.4f}  {: ^16.4f}'.format(mean_value, current_value),
                                                       xy=xy_pos, xycoords='axes fraction',
                                                       xytext=(50, -shift_y), textcoords='offset points'))

    @staticmethod
    def hold():
        plt.show(block=True)

    def get_info(self):
        pass


class PlotDataTracker(object):
    """ A class to track relevant texts for plots corresponding to selected metrics.
    Data buffers and line objects are accessible via the corresponding data_id.

    """

    def __init__(self, data_ids: list):
        self.data_ids = None
        self.data = {}
        self.line_objs = {}
        self.sub_plot_obj = None
        self._validate(data_ids)
        self._configure()

    def _validate(self, data_ids):
        if isinstance(data_ids, list):
            if len(data_ids) > 0:
                self.data_ids = data_ids
            else:
                raise ValueError('data_ids is empty')
        else:
            raise TypeError('data_ids must be a list, received: {}'.format(type(data_ids)))

    def _configure(self):
        for data_id in self.data_ids:
            self.data[data_id] = None
            self.line_objs[data_id] = None



import copy
import logging

from typing import Any, Dict, Optional, Text

logger = logging.getLogger(__name__)


class TaskType(object):
    CLASSIFICATION = 'classification'
    LABELING = 'labeling'
    SCORING = 'scoring'
    REGRESSION = 'regression'
    MULTI_TARGET_CLASSIFICATION = 'multi_target_classification'
    MULTI_TARGET_REGRESSION = 'multi_target_regression'



def load(**kwargs):
    return _load_from_dict(**kwargs)


def _load_from_dict(**kwargs):
    config = {}
    if kwargs:
        config.update(kwargs)
    return ModelConfig(config)


def override_defaults(
        defaults: Optional[Dict[Text, Any]], custom: Optional[Dict[Text, Any]]
) -> Dict[Text, Any]:
    if defaults:
        cfg = copy.deepcopy(defaults)
    else:
        cfg = {}

    if custom:
        if isinstance(custom, dict):
            cfg.update(custom)
        else:
            cfg.update(custom.__dict__)
    return cfg


def component_config_from_pipeline(
        name,
        pipeline,
        defaults=None):
    for c in pipeline:
        if c.get("name") == name:
            return override_defaults(defaults, c)
    else:
        return override_defaults(defaults, {})


class ModelConfig:
    def __init__(self, configuration_values=None):
        """Create a model configuration, optionally overriding
        defaults with a dictionary ``configuration_values``.
        """
        if not configuration_values:
            configuration_values = {}

        self.language = "en"
        self.pipeline = []
        self.data = None

        self.override(configuration_values)

        if self.__dict__["pipeline"] is None:
            # replaces NoneModelConfig with empty list
            self.__dict__["pipeline"] = []
        elif isinstance(self.__dict__["pipeline"], list):

            self.pipeline = [{"name": c} for c in self.__dict__['pipeline']]

            if self.pipeline:
                # replaces the template with the actual components
                self.__dict__["pipeline"] = self.pipeline

        for key, value in self.items():
            setattr(self, key, value)

    def __getitem__(self, key):
        return self.__dict__[key]

    def get(self, key, default=None):
        return self.__dict__.get(key, default)

    def __setitem__(self, key, value):
        self.__dict__[key] = value

    def __delitem__(self, key):
        del self.__dict__[key]

    def __contains__(self, key):
        return key in self.__dict__

    def __len__(self):
        return len(self.__dict__)

    def __getstate__(self):
        return self.as_dict()

    def __setstate__(self, state):
        self.override(state)

    def items(self):
        return list(self.__dict__.items())

    def as_dict(self):
        return dict(list(self.items()))

    def for_component(self, index, defaults=None):
        return component_config_from_pipeline(index, self.pipeline, defaults)

    @property
    def component_names(self):
        if self.pipeline:
            return [c.get("name") for c in self.pipeline]
        else:
            return []

    def override(self, config):
        if config:
            self.__dict__.update(config)



from sklearn.datasets import make_classification


def create_dataset(n_samples=1000, weights=(0.01, 0.01, 0.98), n_classes=3,
                   class_sep=0.8, n_clusters=1):
    return make_classification(n_samples=n_samples, n_features=2,
                               n_informative=2, n_redundant=0, n_repeated=0,
                               n_classes=n_classes,
                               n_clusters_per_class=n_clusters,
                               weights=list(weights),
                               class_sep=class_sep, random_state=0)



import numpy as np


class ConfusionMatrix():
    """ ConfusionMatrix
    This structure constitutes a confusion matrix, or an error matrix. It is
    represented by a matrix of shape (n_labels, n_labels), in a simple, one
    classification task context.
    """

    def __init__(self, n_targets=None, dtype=np.int64):
        super().__init__()
        if n_targets is not None:
            self.n_targets = n_targets
        else:
            self.n_targets = 0
        self.sample_count = 0
        self.dtype = dtype

        self.confusion_matrix = np.zeros((self.n_targets, self.n_targets), dtype=dtype)

    def restart(self, n_targets):
        if n_targets is None:
            self.n_targets = 0
        else:
            self.n_targets = n_targets
        self.confusion_matrix = np.zeros((self.n_targets, self.n_targets))
        self.sample_count = 0
        pass

    def _update(self, i, j):
        self.confusion_matrix[i, j] += 1
        self.sample_count += 1
        return True

    def update(self, i=None, j=None):
        """ update
        Increases by one the count of occurrences in one of the ConfusionMatrix's
        cells.
        Parameters
        ---------
        i: int
            The index of the row to be updated.
        j: int
            The index of the column to be updated.
        Returns
        -------
        bool
            True if the update was successful and False if it was unsuccessful,
            case in which a index is out of range.
        Notes
        -----
        No IndexError or IndexOutOfRange errors raised.
        """
        if i is None or j is None:
            return False

        else:
            m, n = self.confusion_matrix.shape
            if (i <= m) and (i >= 0) and (j <= n) and (j >= 0):
                return self._update(i, j)

            else:
                max_value = np.max(i, j)
                if max_value > m + 1:
                    return False

                else:
                    self.reshape(max_value, max_value)
                    return self._update(i, j)

    def remove(self, i=None, j=None):
        """ remove
        Decreases by one the count of occurrences in one of the ConfusionMatrix's
        cells.
        Parameters
        ----------
        i: int
            The index of the row to be updated.
        j: int
            The index of the column to be updated.
        Returns
        -------
        bool
            True if the removal was successful and False otherwise.
        Notes
        -----
        No IndexError or IndexOutOfRange errors raised.
        """
        if i is None or j is None:
            return False

        m, n = self.confusion_matrix.shape
        if (i <= m) and (i >= 0) and (j <= n) and (j >= 0):
            return self._remove(i, j)

        else:
            return False

    def _remove(self, i, j):
        self.confusion_matrix[i, j] = self.confusion_matrix[i, j] - 1
        self.sample_count -= 1
        return True

    def reshape(self, m, n):
        i, j = self.confusion_matrix.shape

        if (m != n) or (m < i) or (n < j):
            return False
        aux = self.confusion_matrix.copy()
        self.confusion_matrix = np.zeros((m, n), self.dtype)

        for p in range(i):
            for q in range(j):
                self.confusion_matrix[p, q] = aux[p, q]

        return True

    def shape(self):
        """ shape
        Returns
        -------
        tuple
            The confusion matrix's shape.
        """
        return self.confusion_matrix.shape

    def value_at(self, i, j):
        """ value_at
        Parameters
        ----------
        i: int
            An index from one of the matrix's rows.
        j: int
            An index from one of the matrix's columns.
        Returns
        -------
        int
            The current occurrence count at position [i, j].
        """
        return self.confusion_matrix[i, j]

    def row(self, r):
        """ row
        Parameters
        ----------
        r: int
            An index from one of the matrix' rows.
        Returns
        -------
        numpy.array
            The complete row indexed by r.
        """
        return self.confusion_matrix[r: r + 1, :]

    def column(self, c):
        """ column
        Parameters
        ----------
        c: int
            An index from one of the matrix' columns.
        Returns
        -------
        numpy.array
            The complete column indexed by c.
        """
        return self.confusion_matrix[:, c: c + 1]

    def get_sum_main_diagonal(self):
        """ Computes the sum of occurrences in the main diagonal.
        Returns
        -------
        int
            The occurrence count in the main diagonal.
        """
        m, n = self.confusion_matrix.shape
        sum_main_diagonal = 0
        for i in range(m):
            sum_main_diagonal += self.confusion_matrix[i, i]
        return sum_main_diagonal

    @property
    def matrix(self):
        if self.confusion_matrix is not None:
            return self.confusion_matrix
        else:
            return None

    @property
    def _sample_count(self):
        return self.sample_count

    def get_info(self):
        return 'ConfusionMatrix: n_targets: ' + str(self.n_targets) + \
               ' - sample_count: ' + str(self.sample_count) + \
               ' - dtype: ' + str(self.dtype)

    def get_class_type(self):
        return 'collection'


class MOLConfusionMatrix():
    """
    This structure represents a confusion matrix, or an error matrix. It is
    represented by a matrix of shape (n_targets, n_labels, n_labels). It
    basically works as an individual ConfusionMatrix for each of the
    classification tasks in a multi label environment. Thus, n_labels is
    always 2 (binary).

    The first dimension defines which classification task it keeps track of.
    The second dimension is associated with the true y_values, while the other
    is associated with the predictions. For example, an entry in position
    [2, 1, 2] represents a miss classification in the classification task of
    index 2, where the true label was index 1, but the prediction was index 2.

    This structure is used to keep updated statistics from a multi output
    classifier's performance, which allows to compute different evaluation
    metrics.

    """

    def __init__(self, n_targets=None, dtype=np.int64):
        super().__init__()
        if n_targets is not None:
            self.n_targets = n_targets
        else:
            self.n_targets = 0
        self.dtype = dtype
        self.confusion_matrix = np.zeros((self.n_targets, 2, 2), dtype=dtype)

    def restart(self, n_targets):
        if n_targets is None:
            self.n_targets = 0
        else:
            self.n_targets = n_targets
        self.confusion_matrix = np.zeros((self.n_targets, 2, 2), dtype=self.dtype)
        pass

    def _update(self, target, true, pred):
        self.confusion_matrix[int(target), int(true), int(pred)] += 1
        return True

    def update(self, target=None, true=None, pred=None):
        """ update

        Increases by one the occurrence count in one of the matrix's positions.
        As entries arrive, it may reshape the matrix to correctly accommodate all
        possible y_values.

        The count will be increased in the matrix's [label, true, pred] position.

        Parameters
        ----------
        target: int
            A classification task's index.

        true: int
            A true label's index.

        pred: int
            A prediction's index

        Returns
        -------
        bool
            True if the update was successful, False otherwise.

        """

        if target is None or true is None or pred is None:
            return False
        else:
            m, n, p = self.confusion_matrix.shape
            if (target < m) and (target >= 0) and (true < n) and (true >= 0) and (pred < p) and (pred >= 0):
                return self._update(target, true, pred)
            else:
                try:
                    if (true > 1) or (true < 0) or (pred > 1) or (pred < 0):
                        return False
                    if target > m:
                        return False
                    else:
                        self.reshape(target + 1, 2, 2)
                        return self._update(target, true, pred)
                except Exception as e:
                    print(e)

    def remove(self, target=None, true=None, pred=None):
        """ remove

        Decreases by one the occurrence count in one of the matrix's positions.

        The count will be increased in the matrix's [label, true, pred] position.

        Parameters
        ----------
        target: int
            A classification task's index.

        true: int
            A true label's index.

        pred: int
            A prediction's index

        Returns
        -------
        bool
            True if the removal was successful, False otherwise.

        """
        if true is None or pred is None or target is None:
            return False
        m, n, p = self.confusion_matrix.shape
        if (target <= m) and (target >= 0) and (true <= n) and (true >= 0) and (pred >= 0) and (pred <= p):
            return self._remove(target, true, pred)
        else:
            return False

    def _remove(self, target, true, pred):
        self.confusion_matrix[target, true, pred] = self.confusion_matrix[target, true, pred] - 1
        return True

    def reshape(self, target, m, n):
        t, i, j = self.confusion_matrix.shape
        if (target > t + 1) or (m != n) or (m != 2) or (m < i) or (n < j):
            return False
        aux = self.confusion_matrix.copy()
        self.confusion_matrix = np.zeros((target, m, n), self.dtype)
        for w in range(t):
            for p in range(i):
                for q in range(j):
                    self.confusion_matrix[w, p, q] = aux[w, p, q]
        return True

    def shape(self):
        return self.confusion_matrix.shape

    def value_at(self, target, i, j):
        """ value_at

        Parameters
        ----------
        target: int
            An index from one of classification's tasks.

        i: int
            An index from one of the matrix's rows.

        j: int
            An index from one of the matrix's columns.

        Returns
        -------
        int
            The current occurrence count at position [label, i, j].

        """
        return self.confusion_matrix[target, i, j]

    def row(self, r):
        """ row

        Parameters
        ----------
        r: int
            An index from one of the matrix' rows.

        Returns
        -------
        numpy.array
            The complete row indexed by r.

        """
        return self.confusion_matrix[r:r + 1, :]

    def column(self, c):
        """ column

        Parameters
        ----------
        c: int
            An index from one of the matrix' columns.

        Returns
        -------
        numpy.array
            The complete column indexed by c.

        """
        return self.confusion_matrix[:, c:c + 1]

    def target(self, t):
        """ label

        Parameters
        ----------
        t: int
            An index from one of the matrix' label.

        Returns
        -------
        numpy.ndarray
            The complete label indexed by t.

        """
        return self.confusion_matrix[t, :, :]

    def get_sum_main_diagonal(self):
        """ get_sum_main_diagonal

        Computes the sum of occurrences in all the main diagonals.

        Returns
        -------
        int
            The occurrence count in the main diagonals.

        """
        t, m, n = self.confusion_matrix.shape
        sum_main_diagonal = 0
        for i in range(t):
            sum_main_diagonal += self.confusion_matrix[i, 0, 0]
            sum_main_diagonal += self.confusion_matrix[i, 1, 1]
        return sum_main_diagonal

    def get_total_sum(self):
        """ get_total_sum

        Returns
        ------
        int
            The sum of occurrences in the matrix.

        """
        return np.sum(self.confusion_matrix)

    def get_total_discordance(self):
        """ get_total_discordance

        The total discordance is defined as all the occurrences where a miss
        classification was detected. In other words it's the sum of all cells
        indexed by [t, i, j] where i and j are different.

        Returns
        -------
        float
            The total discordance from all label's matrices.

        """
        return self.get_total_sum() - self.get_sum_main_diagonal()

    @property
    def matrix(self):
        if self.confusion_matrix is not None:
            return self.confusion_matrix
        else:
            return None

    def get_info(self):
        return 'MOLConfusionMatrix: n_targets: ' + str(self.n_targets) + \
               ' - total_sum: ' + str(self.get_total_sum()) + \
               ' - total_discordance: ' + str(self.get_total_discordance()) + \
               ' - dtype: ' + str(self.dtype)

    def get_class_type(self):
        return 'collection'


class Deque():
    """ Deque
    A simple buffer used to keep track of a limited number of unitary entries. It
    updates the buffer following a FIFO method, meaning that when the buffer is
    full and a new entry arrives, the oldest entry is pushed out of the queue.
    In theory it keeps track of simple, primitive objects, such as numeric values,
    but in practice it can be used to store any kind of object.
    For this framework the FastBuffer is mainly used to keep track of true y_values
    and predictions in a classification task context, so that we can keep updated
    statistics about the task being executed.
    Parameters
    ----------
    max_size: int
        Maximum size of the queue.
    object_list: list
        An initial list. Optional. If given the queue will be started with the
        values from this list.
    """

    def __init__(self, max_size, object_list=None):
        super().__init__()
        # Default values
        self.current_size = 0
        self.max_size = None
        self.buffer = []

        self.configure(max_size, object_list)

    def get_class_type(self):
        return 'data_structure'

    def configure(self, max_size, object_list):
        self.max_size = max_size
        if isinstance(object_list, list):
            self.add_element(object_list)

    def add_element(self, element_list):
        """ add_element
        Adds a new entry to the buffer. In case there are more elements in the
        element_list parameter than there is free space in the queue, elements
        from the queue are iteratively popped from the queue and appended to
        a list, which in the end is returned.
        """
        if (self.current_size + len(element_list)) <= self.max_size:
            for i in range(len(element_list)):
                self.buffer.append(element_list[i])
            self.current_size += len(element_list)
            return None

        else:
            aux = []
            for element in element_list:
                if self.is_full():
                    aux.append(self.get_next_element())
                self.buffer.append(element)
                self.current_size += 1
            return aux

    def get_next_element(self):
        """ get_next_element
        Pop the head of the queue.
        Returns
        -------
        int or float
            The first element in the queue.
        """
        result = None
        if len(self.buffer) > 0:
            result = self.buffer.pop(0)
            self.current_size -= 1
        return result

    def clear_queue(self):
        self._clear_all()

    def _clear_all(self):
        del self.buffer[:]
        self.buffer = []
        self.current_size = 0
        self.configure(self.max_size, None)

    def is_full(self):
        return self.current_size == self.max_size

    def is_empty(self):
        return self.current_size == 0

    def get_current_size(self):
        return self.current_size

    def peek(self):
        """ peek
        Peek the head of the queue, without removing or altering it.
        Returns
        -------
        int or float
            The head of the queue.
        """
        try:
            return self.buffer[0]
        except IndexError:
            return None

    def get_deque(self):
        return self.buffer


class ComplexDeque():
    """ ComplexDeque

    A complex buffer used to keep track of a limited number of complex entries. It
    updates the buffer following a FIFO method, meaning that when the buffer is
    full and a new entry arrives, the oldest entry is pushed out of the queue.

    We use the term complex entry to specify that each entry is a set of n
    predictions, one for each classification task. This structure is used to keep
    updated statistics from a multi output context.

    Parameters
    ----------
    max_size: int
        Maximum size of the queue.

    width: int
        The width from a complex entry, in other words how many classification
        tasks are there to keep track of.

    Examples
    --------
    It works similarly to the FastBuffer structure, except that it keeps track
    of more than one value per entry. For a complete example, please see
    skmultiflow.evaluation.measure_collection.WindowMultiTargetClassificationMeasurements'
    implementation, where the FastComplexBuffer is used to keep track of the
    MultiOutputLearner's statistics.

    """

    def __init__(self, max_size, width):
        super().__init__()
        # Default values
        self.current_size = 0
        self.max_size = None
        self.width = None
        self.buffer = []

        self.configure(max_size, width)

    def get_class_type(self):
        return 'data_structure'

    def configure(self, max_size, width):
        self.max_size = max_size
        self.width = width

    def add_element(self, element_list):
        """ add_element

        Adds a new entry to the buffer. In case there are more elements in the
        element_list parameter than there is free space in the queue, elements
        from the queue are iteratively popped from the queue and appended to
        a list, which in the end is returned.

        Parameters
        ----------
        element_list: list or numpy.array
            A list with all the elements that are to be added to the queue.

        Returns
        -------
        list
            If no elements need to be popped from the queue to make space for new
            entries there is no return. On the other hand, if elements need to be
            removed, they are added to an auxiliary list, and that list is returned.

        """
        is_list = True
        dim = 1
        if hasattr(element_list, 'ndim'):
            dim = element_list.ndim
        if (dim > 1) or hasattr(element_list[0], 'append'):
            size, width = 0, 0
            if hasattr(element_list, 'append'):
                size, width = len(element_list), len(element_list[0])
            elif hasattr(element_list, 'shape'):
                is_list = False
                size, width = element_list.shape
            self.width = width
            if width != self.width:
                return None
        else:
            size, width = 0, 0
            if hasattr(element_list, 'append'):
                size, width = 1, len(element_list)
            elif hasattr(element_list, 'size'):
                is_list = False
                size, width = 1, element_list.size
            self.width = width
            if width != self.width:
                return None

        if not is_list:
            if size == 1:
                items = [element_list.tolist()]
            else:
                items = element_list.tolist()
        else:
            if size == 1:
                items = [element_list]
            else:
                items = element_list

        if (self.current_size + size) <= self.max_size:
            for i in range(size):
                self.buffer.append(items[i])
            self.current_size += size
            return None
        else:
            aux = []
            for element in items:
                if self.is_full():
                    aux.append(self.get_next_element())
                self.buffer.append(element)
                self.current_size += 1
            return aux

    def get_next_element(self):
        """ get_next_element

        Pop the head of the queue.

        Returns
        -------
        tuple
            The first element of the queue.

        """
        result = None
        if len(self.buffer) > 0:
            result = self.buffer.pop(0)
            self.current_size -= 1
        return result

    def clear_queue(self):
        self._clear_all()

    def _clear_all(self):
        del self.buffer[:]
        self.buffer = []
        self.current_size = 0
        self.configure(self.max_size, None)

    def print_queue(self):
        print(self.buffer)

    def is_full(self):
        return self.current_size == self.max_size

    def is_empty(self):
        return self.current_size == 0

    def get_current_size(self):
        return self.current_size

    def peek(self):
        """ peek

        Peek the head of the queue, without removing or altering it.

        Returns
        -------
        tuple
            The head of the queue.

        """
        try:
            return self.buffer[0]
        except IndexError:
            return None

    def get_queue(self):
        return self.buffer

    def get_info(self):
        return 'FastBuffer: max_size: ' + str(self.max_size) \
               + ' - current_size: ' + str(self.current_size) \
               + ' - width: ' + str(self.width)



from sklearn.feature_extraction.text import HashingVectorizer

from kolibri.features.features import Features
from kolibri.stopwords import get_stop_words


class HashingFeaturizer(Features):
    """Bag of words featurizer

    Creates bag-of-words representation of intent features
    using sklearn's `HashingVectorizer`.
    All tokens which consist only of digits (e.g. 123 and 99
    but not ab12d) will be represented by a single feature."""

    name = "hashing_featurizer"

    provides = ["text_features"]

    requires = ["tokens"]

    defaults = {
        # the parameters are taken from
        # sklearn's CountVectorizer

        # regular expression for tokens
        "token_pattern": r'(?u)\b\w\w+\b',

        # remove accents during the preprocessing step
        "strip_accents": None,  # {'ascii', 'unicode', None}

        # list of stop words
        "stop_words": None,  # string {'en'}, list, or None (default)

        # if convert all characters to lowercase
        "lower-case": False,  # bool
        "use_bigram_model": False
    }

    @classmethod
    def required_packages(cls):
        return ["sklearn"]

    def __init__(self, component_config={}):
        """Construct a new count vectorizer using the sklearn framework."""

        super(HashingFeaturizer, self).__init__(component_config)

        self.use_bigram_model = self.component_config["use_bigram_model"]
        # declare class instance for CountVectorizer
        self.stop_words = get_stop_words('en')

        self.vectorizer = HashingVectorizer(decode_error='ignore', n_features=2 ** 18,
                                            alternate_sign=False, tokenizer=self._identity_tokenizer, lowercase=False,
                                            stop_words=self.stop_words)

    def _identity_tokenizer(self, text):
        return text

    def fit(self, X, y):
        return self

    def transform(self, X):
        return self.vectorizer.transform(X)

    def train(self, training_data, **kwargs):
        [self.process(document) for document in training_data]

    def process(self, document, **kwargs):
        document.vector = self.vectorizer.transform([document.tokens])

    def get_info(self):
        return 'hasing vectorizer'



import unittest

import tests.dnn.text.test_labeling.test_bi_lstm_model as base
from kolibri.dnn.tasks.text.labeling import CNN_LSTM_Model


class TestCNN_LSTM_Model(base.TestBiLSTM_Model):

    @classmethod
    def setUpClass(cls):
        cls.EPOCH_COUNT = 1
        cls.TASK_MODEL_CLASS = CNN_LSTM_Model


if __name__ == "__main__":
    unittest.main()



import collections
import operator
from typing import Dict, Any, Tuple

import numpy as np
import tqdm

from kolibri.indexers.base_indexer import BaseIndexer


class LabelIndexer(BaseIndexer):

    def to_dict(self) -> Dict[str, Any]:
        data = super(LabelIndexer, self).to_dict()
        data['config']['multi_label'] = self.multi_label
        return data

    def __init__(self,
                 multi_label: bool = False,
                 **kwargs: Any) -> None:
        from sklearn.preprocessing import MultiLabelBinarizer
        super(LabelIndexer, self).__init__(**kwargs)
        self.multi_label = multi_label
        self.multi_label_binarizer = MultiLabelBinarizer(self.vocab2idx)

    def build_vocab_generator(self,
                              generators) -> None:
        from sklearn.preprocessing import MultiLabelBinarizer
        if self.vocab2idx:
            return

        vocab2idx: Dict[str, int] = {}
        token2count: Dict[str, int] = {}
        for generator in generators:
            if self.multi_label:
                for _, label in tqdm.tqdm(generator, desc="Preparing classification label vocab dict"):
                    for token in label:
                        count = token2count.get(token, 0)
                        token2count[token] = count + 1
            else:
                for _, label in tqdm.tqdm(generator, desc="Preparing classification label vocab dict"):
                    count = token2count.get(label, 0)
                    token2count[label] = count + 1

        sorted_token2count = sorted(token2count.items(),
                                    key=operator.itemgetter(1),
                                    reverse=True)
        token2count = collections.OrderedDict(sorted_token2count)

        for token, token_count in token2count.items():
            if token not in vocab2idx:
                vocab2idx[token] = len(vocab2idx)
        self.vocab2idx = vocab2idx
        self.idx2vocab = dict([(v, k) for k, v in self.vocab2idx.items()])
        self.multi_label_binarizer = MultiLabelBinarizer(self.vocab2idx)

    def get_tensor_shape(self, batch_size: int, seq_length: int) -> Tuple:
        if self.multi_label:
            return batch_size, len(self.vocab2idx)
        else:
            return (batch_size,)

    def transform(self, samples, **kwargs):
        if self.multi_label:
            sample_tensor = self.multi_label_binarizer.transform(samples)
            return sample_tensor

        sample_tensor = [self.vocab2idx[i] for i in samples]
        return np.array(sample_tensor)

    def inverse_transform(self,
                          labels,
                          *,
                          lengths=None,
                          **kwargs):
        if self.multi_label:
            return self.multi_label_binarizer.inverse_transform(labels)
        else:
            return [self.idx2vocab[i] for i in labels]


if __name__ == "__main__":
    pass



def read_conll(file_path: str, file_type='conll'):
    """
    Read conll format data_file
    Args:
        file_path: path of label file
        text_index: index of text texts, default 0
        label_index: index of label texts, default 1

    Returns:

    """

    data = []
    with open(file_path, 'r', encoding='utf-8') as f:
        for line in f:
            rows = line.strip().split(' ')
            if len(rows) == 1:

                yield data
                data = []

            else:
                data.append(rows)



from sqlalchemy import Table, Column, Integer, String, DateTime, ForeignKey
from sqlalchemy.ext.declarative import declared_attr, declarative_base
from sqlalchemy.orm import relationship
from sqlalchemy.sql import func

from kolibri import settings as constants
from kolibri.document import DocumentBase


class ModelBase(object):
    """
    An augmented base class for SqlAlchemy models.
    """

    @declared_attr
    def __tablename__(cls):
        """
        Return the lowercase class name as the name of the table.
        """
        return cls.__name__.lower()

    id = Column(
        Integer,
        primary_key=True,
        autoincrement=True
    )


Base = declarative_base(cls=ModelBase)

tag_association_table = Table(
    'tag_association',
    Base.metadata,
    Column('tag_id', Integer, ForeignKey('tag.id')),
    Column('document_id', Integer, ForeignKey('document.id'))
)


class Tag(Base):
    """
    A tag that describes a document.
    """

    name = Column(
        String(constants.TAG_NAME_MAX_LENGTH),
        unique=True
    )


class Document(Base, DocumentBase):
    """
    A Document represents a sentence or phrase.
    """

    confidence = 0

    text = Column(
        String(constants.DOCUMENT_TEXT_MAX_LENGTH)
    )

    search_text = Column(
        String(constants.DOCUMENT_TEXT_MAX_LENGTH),
        nullable=False,
        server_default=''
    )

    label = Column(
        String(constants.DOCUMENT_LABEL_MAX_LENGTH),
        nullable=False,
        server_default=''
    )

    created_at = Column(
        DateTime(timezone=True),
        server_default=func.now()
    )
    target_text = Column(
        String(constants.DOCUMENT_TEXT_MAX_LENGTH),
        nullable=True
    )

    search_target_text = Column(
        String(constants.DOCUMENT_TEXT_MAX_LENGTH),
        nullable=False,
        server_default=''
    )

    tags = relationship(
        'Tag',
        secondary=lambda: tag_association_table,
        backref='documents'
    )

    def get_tags(self):
        """
        Return a list of tags for this document.
        """
        return [tag.name for tag in self.tags]

    def add_tags(self, *tags):
        """
        Add a list of strings to the document as tags.
        """
        self.tags.extend([
            Tag(name=tag) for tag in tags
        ])



import numpy as np
from tensorflow.keras.layers import Input, Conv1D, GlobalMaxPool1D, Dense, MaxPooling1D, Flatten, Reshape, UpSampling1D
from tensorflow.keras.models import Sequential, Model

from kolibri.dnn.autoencoder.base_autoencoder import BaseAutoencoder


class Conv1DAutoencoder(BaseAutoencoder):

    @classmethod
    def get_default_hyper_parameters(cls):
        return {
            'conv1d_layer': {
                'filters': 128,
                'kernel_size': [5],
                'activation': 'relu'
            },
            'max_pool_layer': {},
            'dense_layer': {
                'units': 64,
                'activation': 'relu'
            },
            'activation_layer': {
                'activation': 'softmax'
            },
            'multipooling': False
        }

    def build_model_arc(self):
        NUM_WORDS = 3
        pool_size = 2
        x = Input(shape=(32, 100), name="input")
        h = x
        h = Conv1D(filters=100, kernel_size=NUM_WORDS,
                   activation="relu", padding='same', name='Conv1')(h)
        h = MaxPooling1D(pool_size=pool_size, name='Maxpool1')(h)
        h = Conv1D(filters=50, kernel_size=NUM_WORDS,
                   activation="relu", padding='same', name='Conv2')(h)
        h = MaxPooling1D(pool_size=pool_size, name="Maxpool2")(h)
        h = Flatten()(h)
        h = Dense(8, name='embedding')(h)
        y = h
        y = Dense(1050, activation="relu")(y)
        y = Reshape((8, 150))(y)
        y = Conv1D(filters=50, kernel_size=NUM_WORDS,
                   activation="relu", padding='same', name='conv-decode1')(y)
        y = UpSampling1D(size=pool_size, name='upsampling1')(y)
        y = Conv1D(filters=100, kernel_size=NUM_WORDS,
                   activation="relu", padding='same', name='conv-decode2')(y)
        y = UpSampling1D(size=pool_size, name='upsampling2')(y)

        autoencoderM = Model(x, y)
        myLoss = 'mean_squared_error'
        autoencoderM.compile(optimizer='adadelta', loss=myLoss)

        autoencoderM.summary()  # will print





        self.autoencoder = Sequential()
        self.autoencoder.add(Conv1D(filters=256, kernel_size=5, padding='same', activation='relu',
                         input_shape=(self.input_dim, 1)))
        self.autoencoder.add(GlobalMaxPool1D())

        self.autoencoder.add(Dense(units=self.input_dim, activation='linear'))

    def fit(self, x_train, x_validate=None,batch_size=32, epochs=10,callbacks=None, fit_kwargs=None, shuffle= True):
        """

        """
        self.build_model(x_train)

        return self.autoencoder.fit(np.expand_dims(x_train, axis=2), x_train,
                                        epochs=epochs,
                                        validation_data=(x_validate, x_validate),
#                                        callbacks=callbacks
                                        )


    def predict(self,
               x_data,
               batch_size=32,
               debug_info=False,
               predict_kwargs= None):
        """
        Generates output predictions for the encoder.
        """
        if predict_kwargs is None:
            predict_kwargs = {}
        pred = self.autoencoder.predict(np.expand_dims(x_data,axis=2), batch_size=batch_size, **predict_kwargs)

        return pred


import unittest

from kolibri.data.corpus.corpus import CONLL2003ENCorpus
from kolibri.data.corpus.corpus import ConsumerComplaintsCorpus


class TestConsumerComplaintsCorpus(unittest.TestCase):

    def test_load_data(self):
        corpus = ConsumerComplaintsCorpus()
        train_x, train_y = corpus.get_data()
        assert len(train_x) == len(train_y)
        assert len(train_x) > 0
        assert train_x[:5] != train_y[:5]
        corpus = ConsumerComplaintsCorpus('test')
        test_x, test_y = corpus.get_data('')
        assert len(test_x) == len(test_y)
        assert len(test_x) > 0

        corpus = ConsumerComplaintsCorpus('train')
        test_x, test_y = corpus.get_data('')
        assert len(test_x) == len(test_y)
        assert len(test_x) > 0

        corpus = ConsumerComplaintsCorpus('validate')
        test_x, test_y = corpus.get_data('')
        assert len(test_x) == len(test_y)
        assert len(test_x) > 0


class TestCONLL2003ENCorpus(unittest.TestCase):

    def test_load_data(self):
        corpus = CONLL2003ENCorpus('train')
        train_x, train_y = corpus.get_data()

        assert len(train_x) == len(train_y)
        assert len(train_x) > 0
        assert train_x[:5] != train_y[:5]

        corpus = CONLL2003ENCorpus('test')
        test_x, test_y = corpus.get_data()
        assert len(test_x) == len(test_y)
        assert len(test_x) > 0

        corpus = CONLL2003ENCorpus('valid')
        test_x, test_y = corpus.get_data()
        assert len(test_x) == len(test_y)
        assert len(test_x) > 0


if __name__ == "__main__":
    pass



from strsimpy.levenshtein import Levenshtein
from strsimpy.normalized_levenshtein import NormalizedLevenshtein

from kolibri.similarity.distance import Distance

levenshtein = Levenshtein()
normalised_levenshtein = NormalizedLevenshtein()


class LevenshteinDistance(Distance):

    def __init__(self, normalize=True):
        self.normalize = normalize

    def compare(self, string1, string2):
        if self.normalize:
            return normalised_levenshtein.distance(string1, string2)
        return levenshtein.distance(string1, string2)



import warnings

import numpy as np
import pandas as pd

from kolibri.data.corpus.base_stream import BaseStream


class DataStream(BaseStream):
    """ Creates a stream from a texts source.

    DataStream takes the whole texts set containing the `X` (features) and `Y` (targets) or takes `X` and `Y` separately.
    For the first case `target_idx` and `n_targets` need to be provided, in the second case they are not needed.
    """

    _CLASSIFICATION = 'classification'
    _REGRESSION = 'regression'
    _Y_is_defined = False

    def __init__(self, data, y=None, target_idx=-1, n_targets=1, cat_features=None, name=None, allow_nan=False):
        super().__init__()
        self.X = None
        self.y = y
        self.cat_features = cat_features
        self.cat_features_idx = [] if self.cat_features is None else self.cat_features
        self.n_targets = n_targets
        self.target_idx = target_idx
        self.task_type = None
        self.n_classes = 0
        self.data = data
        self._is_ready = False
        self.name = name
        self.allow_nan = allow_nan

        if self._Y_is_defined:
            self.y = pd.DataFrame(self.y)
            if self.y.shape[0] != self.data.shape[0]:
                raise ValueError("X and y should have the same number of rows")
            else:
                if isinstance(self.data, pd.DataFrame):
                    self.X=self.data
                else:
                    self.X = pd.DataFrame(self.data)
                self.target_idx = -self.y.shape[1]
                self.n_targets = self.y.shape[1]

        self.prepare()

    @property
    def y(self):
        """
        Return the targets' columns.

        Returns
        -------
        np.ndarray:
            the targets' columns
        """
        return self._y

    @y.setter
    def y(self, y):
        """
        Sets the targets' columns

        Parameters
        ----------
        y: pd.DataFrame or np.ndarray
            the targets' columns

        """
        if y is not None and not self._Y_is_defined:
            self._Y_is_defined = True
        if not self._Y_is_defined or (isinstance(y, np.ndarray) or isinstance(y, pd.DataFrame)):
            self._y = y
        else:
            raise ValueError("np.ndarray or pd.DataFrame y object expected, and {} was passed".format(type(y)))

    @property
    def X(self):
        """
        Return the features' columns.

        Returns
        -------
        np.ndarray:
            the features' columns
        """
        return self._X

    @X.setter
    def X(self, X):
        """
        Sets the features' columns.

        Parameters
        ----------
        X: pd.DataFrame or np.ndarray
            the features' columns.
        """

        if isinstance(X, np.ndarray) or isinstance(X, pd.DataFrame) or not self._Y_is_defined:
            self._X = X

        else:
            raise ValueError("np.ndarray or pd.DataFrame X object expected, and {} was passed".format(type(X)))

    @property
    def data(self):
        """
        Return the texts set used to generate the stream.

        Returns
        -------
        pd.DataFrame:
            Data set.
        """
        return self._data

    @data.setter
    def data(self, data):
        """
        Sets the texts set used to generate the stream.

        Parameters
        ----------
        data: DataFrame or np.ndarray
            the texts set

        """

        if isinstance(data, pd.DataFrame):
            self._data = data
        elif isinstance(data, np.ndarray):
            self._data = pd.DataFrame(data)
        else:
            raise ValueError("Invalid type {}, for texts".format(type(data)))

    @data.deleter
    def data(self):
        """
            Deletes texts
        """
        del self._data

    @property
    def target_idx(self):
        """
        Get the number of the column where Y begins.

        Returns
        -------
        int:
            The number of the column where Y begins.
        """
        return self._target_idx

    @target_idx.setter
    def target_idx(self, target_idx):
        """
        Sets the number of the column where Y begins.

        Parameters
        ----------
        target_idx: int
        """

        self._target_idx = target_idx

    @property
    def n_targets(self):
        """
         Get the number of targets.

        Returns
        -------
        int:
            The number of targets.
        """
        return self._n_targets

    @n_targets.setter
    def n_targets(self, n_targets):
        """
        Sets the number of targets.

        Parameters
        ----------
        n_targets: int
        """

        self._n_targets = n_targets

    @property
    def cat_features_idx(self):
        """
        Get the list of the categorical features index.

        Returns
        -------
        list:
            List of categorical features index.

        """
        return self._cat_features_idx

    @cat_features_idx.setter
    def cat_features_idx(self, cat_features_idx):
        """
        Sets the list of the categorical features index.

        Parameters
        ----------
        cat_features_idx:
            List of categorical features index.
        """

        self._cat_features_idx = cat_features_idx

    def prepare(self):
        self.restart()
        if not self._is_ready:
            if self._Y_is_defined:
                self._load_X_y()
            else:
                self._load_data()
                del self.data
            self._is_ready = True

    def _load_X_y(self):

        self.y = pd.DataFrame(self.y)

        check_data_consistency(self.y, self.allow_nan)
        check_data_consistency(self.X, self.allow_nan)

        self.n_samples, self.n_features = self.X.shape
        self.feature_names = self.X.columns.values.tolist()
        self.target_names = self.y.columns.values.tolist()

        self.y = self.y.values
        self.X = self.X.values
        self.n_cat_features=0
        if self.cat_features_idx:
            if max(self.cat_features_idx) < self.n_features:
                self.n_cat_features = len(self.cat_features_idx)
            else:
                raise IndexError('Categorical feature index in {} '
                                 'exceeds n_features {}'.format(self.cat_features_idx, self.n_features))
        self.n_num_features = self.n_features - self.n_cat_features

        if np.issubdtype(self.y.dtype, np.integer):
            self.task_type = self._CLASSIFICATION
            self.n_classes = len(np.unique(self.y))
        else:
            self.task_type = self._REGRESSION

        self.target_values = self._get_target_values()

    def _load_data(self):

        check_data_consistency(self.data, self.allow_nan)

        rows, cols = self.data.shape
        self.n_samples = rows
        labels = self.data.columns.values.tolist()

        if (self.target_idx + self.n_targets) == cols or (self.target_idx + self.n_targets) == 0:
            # Take everything to the right of target_idx
            self.y = self.data.iloc[:, self.target_idx:].values
            self.target_names = self.data.iloc[:, self.target_idx:].columns.values.tolist()
        else:
            # Take only n_targets columns to the right of target_idx, use the rest as features
            self.y = self.data.iloc[:, self.target_idx:self.target_idx + self.n_targets].values
            self.target_names = labels[self.target_idx:self.target_idx + self.n_targets]

        self.X = self.data.drop(self.target_names, axis=1).values
        self.feature_names = self.data.drop(self.target_names, axis=1).columns.values.tolist()

        _, self.n_features = self.X.shape
        if self.cat_features_idx:
            if max(self.cat_features_idx) < self.n_features:
                self.n_cat_features = len(self.cat_features_idx)
            else:
                raise IndexError('Categorical feature index in {} '
                                 'exceeds n_features {}'.format(self.cat_features_idx, self.n_features))
        self.n_num_features = self.n_features - self.n_cat_features

        if np.issubdtype(self.y.dtype, np.integer):
            self.task_type = self._CLASSIFICATION
            self.n_classes = len(np.unique(self.y))
        else:
            self.task_type = self._REGRESSION

        self.target_values = self._get_target_values()

    def restart(self):
        """ Restarts the stream.

        It basically server the purpose of reinitializing the stream to
        its initial state.

        """
        self.sample_idx = 0
        self.current_sample_x = None
        self.current_sample_y = None

    def next(self, batch_size=1):
        """ Returns next sample from the stream.

        If there is enough instances to supply at least batch_size samples, those
        are returned. If there aren't a tuple of (None, None) is returned.

        Parameters
        ----------
        batch_size: int (optional, default=1)
            The number of instances to return.

        Returns
        -------
        tuple or tuple list
            Returns the next batch_size instances.
            For general purposes the return can be treated as a numpy.ndarray.

        """
        self.sample_idx += batch_size
        try:

            self.current_sample_x = self.X[self.sample_idx - batch_size:self.sample_idx, :]
            self.current_sample_y = self.y[self.sample_idx - batch_size:self.sample_idx, :]
            if self.n_targets < 2:
                self.current_sample_y = self.current_sample_y.flatten()

        except IndexError:
            self.current_sample_x = None
            self.current_sample_y = None
        return self.current_sample_x, self.current_sample_y

    def has_more_samples(self):
        """ Checks if stream has more samples.

        Returns
        -------
        Boolean
            True if stream has more samples.

        """
        return (self.n_samples - self.sample_idx) > 0

    def n_remaining_samples(self):
        """ Returns the estimated number of remaining samples.

        Returns
        -------
        int
            Remaining number of samples.

        """
        return self.n_samples - self.sample_idx

    def print_df(self):
        """
        Prints all the samples in the stream.

        """
        print(self.X)
        print(self.y)

    def get_data_info(self):
        name = self.name + ": " if self.name else ""
        if self.task_type == self._CLASSIFICATION:
            return "{}{} label(s), {} classes".format(name, self.n_targets, self.n_classes)
        elif self.task_type == self._REGRESSION:
            return "{} label(s)".format(name, self.n_targets)

    def _get_target_values(self):
        if self.task_type == 'classification':
            if self.n_targets == 1:
                return np.unique(self.y).tolist()
            else:
                return [np.unique(self.y[:, i]).tolist() for i in range(self.n_targets)]
        elif self.task_type == self._REGRESSION:
            return [float] * self.n_targets

    def get_info(self):
        return 'DataStream(n_targets={}, target_idx={}, cat_features={}, name={})'. \
            format(self.target_idx, self.n_targets, self.cat_features,
                   self.name if not self.name else "'" + self.name + "'")


def check_data_consistency(raw_data_frame, allow_nan=False):
    """
    Check texts consistency with respect to scikit-multiflow assumptions:

    * Only numeric texts types are used.
    * Missing values are, in general, not supported.

    Parameters
    ----------
    raw_data_frame: pandas.DataFrame
        The texts frame containing the texts to check.

    allow_nan: bool, optional (default=False)
        If True, allows NaN values in the texts. Otherwise, an error is raised.

    """
    if (raw_data_frame.dtypes == 'object').values.any():
        raise ValueError('Non-numeric texts found:\n {}'
                         'scikit-multiflow only supports numeric texts.'.format(raw_data_frame.dtypes))

    if raw_data_frame.isnull().values.any():
        if not allow_nan:
            raise ValueError("NaN values found. Missing values are not fully supported.\n"
                             "You can deactivate this error via the 'allow_nan' option.")
        else:
            warnings.warn("NaN values found. Functionality is not guaranteed for some methods. Proceed with caution.",
                          UserWarning)



import functools
import operator

import numpy as np


def get_shape(l):
    d = 0
    while isinstance(l, list):
        d += 1
        l = l[0]
    return d


def flatten(l):
    while get_shape(l) >= 2:
        l = functools.reduce(operator.iconcat, l, [])
    return l


def chunks(lst, n):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i:i + n]


def get_dimensions(X):
    """ Return the dimensions from a numpy.array, numpy.ndarray or list.

    Parameters
    ----------
    X: numpy.array, numpy.ndarray, list, list of lists.

    Returns
    -------
    tuple
        A tuple representing the X structure's dimensions.
    """
    r, c = 1, 1
    if isinstance(X, type(np.array([0]))):
        if X.ndim > 1:
            r, c = X.shape
        else:
            r, c = 1, X.size

    elif isinstance(X, type([])):
        if isinstance(X[0], type([])):
            r, c = len(X), len(X[0])
        else:
            c = len(X)

    return r, c



import logging
import os

import joblib
import numpy as np
from scipy.sparse import vstack
from sklearn.utils import class_weight

from kolibri.indexers.label_indexer import LabelIndexer
from kolibri.kolibri_component import Component

logger = logging.getLogger(__name__)

KOLIBRI_MODEL_FILE_NAME = "classifier_kolibri.pkl"
DNN_MODEL_FILE_NAME = "classifier_dnn"


class DnnEstimator(Component):
    """classifier using the sklearn framework"""

    _estimator_type = 'estimator'

    name = ''

    provides = []

    requires = []

    defaults = {

        # the models used in the classifier if several models are given they will be combined
        "embeddings": None,
        "multi-label": False,
        "sequence_length": 'auto',
        "epochs": 1,
        "loss": 'categorical_crossentropy',
        "class-weight": False,
    }

    def __init__(self, component_config=None):

        """Construct a new class classifier using the sklearn framework."""
        super().__init__(component_config=component_config)

        self.indexer = LabelIndexer(multi_label=self.component_config["multi-label"])

    @classmethod
    def required_packages(cls):
        return ["tensorflow"]

    def fit(self, X, y, X_val=None, y_val=None):
        fit_kwargs = {}
        if self.component_config['class-weight']:
            class_weights = class_weight.compute_class_weight('balanced',
                                                              np.unique(y),
                                                              y)
            fit_kwargs = {"class_weight": class_weights}

        self.clf.fit(X, y, x_validate=X_val, y_validate=y_val, epochs=self.component_config["epochs"],
                     fit_kwargs=fit_kwargs)

    def transform(self, document):

        return self.clf.transform(document, )

    def predict(self, X):
        """Given a bow vector of an input text, predict most probable label.

        Return only the most likely label.

        :param X: bow of input text
        :return: tuple of first, the most probable label and second,
                 its probability."""

        return self.clf.predict(X)

    def train(self, training_data, **kwargs):

        y = [document.label for document in training_data]
        X = vstack([document.vector for document in training_data])
        self.fit(X, y)

    def process(self, document, **kwargs):
        raise NotImplementedError

    def __getstate__(self):
        """Return state values to be pickled."""
        return (self.component_config, self.classifier_type, self.indexer)

    def __setstate__(self, state):
        """Restore state from the unpickled state values."""
        self.component_config, self.classifier_type, self.indexer = state

    @classmethod
    def load(cls, model_dir=None, model_metadata=None, cached_component=None, **kwargs):

        meta = model_metadata.for_component(cls.name)
        classifier_file_name = meta.get("classifier_file", KOLIBRI_MODEL_FILE_NAME)
        dnn_file_name = meta.get("dnn_file", DNN_MODEL_FILE_NAME)
        classifier_file = os.path.join(model_dir, classifier_file_name)
        if os.path.exists(classifier_file):
            # Load saved model
            model = joblib.load(classifier_file)

            clf = model.classifier_type.load_model(os.path.join(model_dir, dnn_file_name))

            model.clf = clf
            return model
        else:
            return cls(meta)

    def persist(self, model_dir):
        """Persist this model into the passed directory.

        Returns the metadata necessary to load the model again."""
        classifier_file = os.path.join(model_dir, KOLIBRI_MODEL_FILE_NAME)
        joblib.dump(self, classifier_file)
        dnn_file = os.path.join(model_dir, DNN_MODEL_FILE_NAME)
        if self.clf:
            self.clf.save(dnn_file)

        return {"classifier_file": KOLIBRI_MODEL_FILE_NAME, "dnn_file": DNN_MODEL_FILE_NAME}

    #
    # @classmethod
    # def load(cls, model_dir=None, model_metadata=None,  cached_component=None,  **kwargs):
    #
    #
    #     meta = model_metadata.for_component(cls.name)
    #     classifier_file_name = meta.get("classifier_file", KOLIBRI_MODEL_FILE_NAME)
    #     dnn_file_name = meta.get("dnn_file", DNN_MODEL_FILE_NAME)
    #     classifier_file = os.path.join(model_dir, classifier_file_name)
    #
    #     if os.path.exists(classifier_file):
    #         # Load saved model
    #         model = joblib.load(classifier_file)
    #         clf = kolibri.dnn.utils.load_model(dnn_file_name)
    #         model.clf=clf
    #         return model
    #     else:
    #         return cls(meta)
    #
    # def persist(self, model_dir):
    #     """Persist this model into the passed directory."""
    #
    #
    #
    #     classifier_file = os.path.join(model_dir, KOLIBRI_MODEL_FILE_NAME)
    #     joblib.dump(self, classifier_file)
    #     dnn_file = os.path.join(model_dir, DNN_MODEL_FILE_NAME)
    #     self.clf.save(dnn_file)
    #
    #     return {"classifier_file": DNN_MODEL_FILE_NAME, "dnn_file":DNN_MODEL_FILE_NAME}



import csv


def read_csv(file_name, file_type='csv', separator=',', encoding='utf-8'):
    if file_type in ['csv', 'txt', 'tsv']:
        with open(file_name, encoding=encoding) as file:
            reader = csv.DictReader(file, delimiter=separator)
            for line in reader:
                yield line
    else:
        raise NotImplementedError("file type not yet supported")



# -*- coding: utf-8 -*-

import itertools
from typing import Optional, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from kolibri.classifier_evaluator.metrics import confusion_matrix
from kolibri.classifier_evaluator.pre_process import data_type_converter

# rc('text', usetex=True)  # use latext
# Setting chinese encoding:
# plt.rcParams['font.sans-serif']=['Microsoft YaHei']  # plt.rcParams['font.sans-serif']=['SimHei']
# plt.rcParams['axes.unicode_minus'] = False
plt.rcParams.update({'font.size': 6})

_DEFAULT_FIGURE_SIZE = (10, 10)
_DEFAULT_COLORMAP = {'diagonal': 'Greens', 'non_diagonal': 'Oranges', 'summary': 'Blues'}


def plot_confusion_matrix_by_dict(confusion_matrix_dict: dict,
                                  class_order: Optional[list] = None,
                                  **kwargs):
    """
    plot a confusion matrix heat map;
    :param confusion_matrix_dict: dict, a confusion matrix dict, e.g.,
        {('A', 'A'): 0, ('A', 'B'): 1, ...}, the key being ('true class', 'predicted class'), and the value being
        the count of the instances with the true class and predicted class;
    :param class_order: [list, None], if specify, confusion matrix's axes will follow the order given;
    :param kwargs:
        :param colormap: dict, that specify the color map for diagonal, non-diagonal, and summary elements in the matrix
            plot;
        :param figure_size: tuple, e.g., (10, 10), the size of the figure;
        :param subtitle: str, sub title given to the plot;
        :param save: [bool, str], if it is saved to an external .png file;
    :return: plt.show or saved image;
    """

    if 'colormap' in kwargs:
        colormap = kwargs['colormap']
    else:
        colormap = _DEFAULT_COLORMAP

    if 'figure_size' in kwargs:
        figure_size = kwargs['figure_size']
    else:
        figure_size = _DEFAULT_FIGURE_SIZE

    if not class_order:
        class_order = list(set([key[0] for key in confusion_matrix_dict.keys()]))

    # unscaled confusion matrix:
    confusion_matrix = np.ndarray(shape=[len(class_order), len(class_order)], dtype=float)
    for index_true, true_class in enumerate(class_order):
        for index_predicted, predicted_class in enumerate(class_order):
            confusion_matrix[index_true, index_predicted] = \
                confusion_matrix_dict[(true_class, predicted_class)]

    confusion_matrix = np.vstack([confusion_matrix, confusion_matrix.sum(axis=0)])
    confusion_matrix = np.hstack([confusion_matrix, confusion_matrix.sum(axis=1).reshape(confusion_matrix.shape[0], 1)])

    # confusion matrix scaled by true label:
    confusion_matrix_nor_true = np.divide(confusion_matrix, confusion_matrix[:, -1][:, None],
                                          out=np.zeros_like(confusion_matrix),
                                          where=confusion_matrix[:, -1][:, None] != 0)
    # adjust the last row to be fraction of total population;
    confusion_matrix_nor_true[:, -1] = np.divide(confusion_matrix[:, -1], confusion_matrix[-1, -1],
                                                 out=np.zeros_like(confusion_matrix[:, -1]),
                                                 where=confusion_matrix[-1, -1] != 0)

    # confusion matrix scaled by predicted label:
    confusion_matrix_nor_predicted = np.divide(confusion_matrix, confusion_matrix[-1, :][None, :],
                                               out=np.zeros_like(confusion_matrix),
                                               where=confusion_matrix[-1, :][None, :] != 0)
    # adjust the last row to be fraction of total population;
    confusion_matrix_nor_predicted[-1, :] = np.divide(confusion_matrix[-1, :], confusion_matrix[-1, -1],
                                                      out=np.zeros_like(confusion_matrix[-1, :]),
                                                      where=confusion_matrix[-1, -1] != 0)

    # col names:
    col_names = class_order + ['sum']

    plt.figure(figsize=figure_size)
    cmap_base = plt.get_cmap(colormap['non_diagonal'])
    rgba_base = cmap_base(confusion_matrix_nor_true)

    # replace diagonal in the confusion matrix to different color spectrum;
    cmap_diagonal = plt.get_cmap(colormap['diagonal'])
    rbga_diagonal = cmap_diagonal(np.diag(confusion_matrix_nor_true))
    for diagonal_index, col in enumerate(rbga_diagonal):
        rgba_base[diagonal_index, diagonal_index] = col

    # replace the true column to different color spectrum:
    cmap_sum_true = plt.get_cmap(colormap['summary'])
    rbga_sum_true = cmap_sum_true(confusion_matrix_nor_true[:, -1])
    for index, col in enumerate(rbga_sum_true):
        rgba_base[index, -1] = col

    # replace the true column to different color spectrum:
    cmap_sum_predicted = plt.get_cmap(colormap['summary'])
    rbga_sum_predicted = cmap_sum_predicted(confusion_matrix_nor_predicted[-1, :])
    for index, col in enumerate(rbga_sum_predicted):
        rgba_base[-1, index] = col

    plt.imshow(rgba_base, interpolation='nearest')
    tick_marks = np.arange(len(col_names))
    plt.xticks(tick_marks, col_names)
    plt.yticks(tick_marks, col_names)
    plt.ylim(len(col_names) - 0.5, -0.5)  # adjusted for the bug before Matplotlib 3.1.2.

    # add annotation text:
    thresh = confusion_matrix_nor_true.max() / 2
    for i, j in itertools.product(range(confusion_matrix.shape[0] - 1),  # left out the last element;
                                  range(confusion_matrix.shape[1] - 1)):

        if i == j:  # annotation for instance with true == predicted;
            num_annotation = r"$\bf{" + str(int(confusion_matrix[i, j])) + "}$" + "\n"

            recall_num = str(round(confusion_matrix_nor_true[i, j], 3))
            recall_annotation = recall_num + r"($\mathit{Recall}$)" + '\n'

            precision_num = str(round(confusion_matrix_nor_predicted[i, j], 3))
            precision_annotation = precision_num + r"($\mathit{Precision}$)"

            annotation_text = num_annotation + recall_annotation + precision_annotation

            plt.text(j, i, annotation_text,
                     horizontalalignment="center",
                     verticalalignment="center",
                     color="white" if confusion_matrix_nor_true[i, j] > thresh else "black")

        else:
            num_annotation = r"$\bf{" + str(int(confusion_matrix[i, j])) + "}$" + "\n"

            recall_num = str(round(confusion_matrix_nor_true[i, j], 3))
            recall_annotation = recall_num + r"($\mathit{/True}$)" + '\n'

            precision_num = str(round(confusion_matrix_nor_predicted[i, j], 3))
            precision_annotation = precision_num + r"($\mathit{/Predicted}$)"

            annotation_text = num_annotation + recall_annotation + precision_annotation

            plt.text(j, i, annotation_text,
                     horizontalalignment="center",
                     verticalalignment="center",
                     color="white" if confusion_matrix_nor_true[i, j] > thresh else "black")

    # add text for the last col (sum):
    last_index = confusion_matrix.shape[0] - 1
    for i in range(confusion_matrix.shape[0]):
        num_annotation = r"$\bf{" + str(int(confusion_matrix[i, -1])) + "}$" + "\n"
        percentage_annotation = r"$\bf{" + str(round(confusion_matrix_nor_true[i, -1] * 100, 2)) + r"\%}$"

        annotation_text = num_annotation + percentage_annotation
        plt.text(last_index, i, annotation_text,
                 horizontalalignment="center",
                 verticalalignment="center",

                 color="white" if confusion_matrix_nor_true[i, -1] > thresh else "black")

    for j in range(confusion_matrix.shape[1]):
        num_annotation = r"$\bf{" + str(int(confusion_matrix[-1, j])) + "}$" + "\n"
        percentage_annotation = r"$\bf{" + str(round(confusion_matrix_nor_predicted[-1, j] * 100, 2)) + r"\%}$"

        annotation_text = num_annotation + percentage_annotation
        plt.text(j, last_index, annotation_text,
                 horizontalalignment="center",
                 verticalalignment="center",
                 color="white" if confusion_matrix_nor_predicted[-1, j] > thresh else "black")

    # plt.tight_layout()
    if 'subtitle' in kwargs:
        title_text = f"Confusion Matrix Plot ({kwargs['subtitle']})"
        saved_file_name = f"confusion_matrix_plot_{kwargs['subtitle']}.png"
    else:
        title_text = f"Confusion Matrix Plot"
        saved_file_name = f"confusion_matrix_plot.png"

    plt.title(title_text)
    plt.ylabel('True label')
    plt.xlabel('Predicted label')

    if 'save' in kwargs:
        save = kwargs['save']
        if isinstance(save, str):
            plt.savefig(save)
        else:
            plt.savefig(saved_file_name)

    plt.show()


def plot_confusion_matrix_on_predicted_series(true_series: Union[pd.Series, np.ndarray, list],
                                              predicted_series: Union[pd.Series, np.ndarray, list],
                                              class_order: Optional[list] = None,
                                              **kwargs):
    """
    run the above confusion plot function after running confusion matrix function;
    :param true_series: [pd.Series, np.ndarray, list, None], a series of true classes;
    :param predicted_series: [pd.Series, np.ndarray, list, None], a series of predicted classes;
    :param class_order: refer class_order param above;
    :param kwargs: refer kwargs above;
    :return: plt.show or saved image;
    """
    true_series = data_type_converter(series=true_series)
    predicted_series = data_type_converter(series=predicted_series)
    confusion_matrix_dict = confusion_matrix(true=true_series,
                                             predicted=predicted_series, normalize=False)

    plot_confusion_matrix_by_dict(confusion_matrix_dict=confusion_matrix_dict,
                                  class_order=class_order, **kwargs)



import json
import os
import threading
from typing import Dict, Any

import numpy as np
import tensorflow as tf
from sklearn import metrics as sklearn_metrics

import kolibri
from kolibri.data.corpus.generators import DataGenerator
from kolibri.dnn.embeddings import DefaultEmbedding
from kolibri.dnn.layers import L
from kolibri.dnn.tasks.base_model import TaskBaseModel
from kolibri.dnn.utils import load_data_object
from kolibri.logger import logger
from kolibri.metrics.multi_label_classification import multi_label_classification_report

threading._DummyThread._Thread__stop = lambda x: 42


class BaseAudioClassificationModel(TaskBaseModel):
    """
    Abstract Audio Classification Model
    """

    __task__ = 'classification'

    def to_dict(self) -> Dict:
        info = super(BaseAudioClassificationModel, self).to_dict()
        info['config']['multi_label'] = self.multi_label
        return info

    def __init__(self, sequence_length=None, hyper_parameters=None, multi_label: bool = False, label_indexer=None):
        """

        Args:
            embedding: embedding object
            sequence_length: target sequence length_train
            hyper_parameters: hyper_parameters to overwrite
            multi_label: is multi-label classification
            label_indexer: label processor
        """
        super().__init__(sequence_length, hyper_parameters, multi_label, label_indexer)

    def _activation_layer(self) -> L.Layer:
        if self.multi_label:
            return L.Activation('sigmoid')
        else:
            return L.Activation('softmax')

    def build_model(self,
                    x_train,
                    y_train):
        """
        Build Model with x_data and y_data

        This function will setup a :class:`CorpusGenerator`,
         then call py:meth:`BaseTextClassificationModel.build_model_gen` for preparing processor and model

        Args:
            x_train:
            y_train:

        Returns:

        """

        train_gen = DataGenerator(x_train, y_train)
        self.build_model_generator([train_gen])

    def build_model_generator(self, generators):
        self.label_indexer.build_vocab_generator(generators)

        if self.tf_model is None:
            self.build_model_arc()
            self.compile_model()

    @classmethod
    def get_default_hyper_parameters(cls) -> Dict[str, Dict[str, Any]]:
        raise NotImplementedError

    def build_model_arc(self) -> None:
        raise NotImplementedError

    def compile_model(self,
                      loss: Any = None,
                      optimizer: Any = None,
                      metrics: Any = None,
                      **kwargs: Any) -> None:
        """
        Configures the model for training.
        call :meth:`tf.keras.Model.predict` to compile model with custom loss, optimizer and metrics
        """
        if loss is None:
            if self.multi_label:
                loss = 'binary_crossentropy'
            else:
                loss = 'sparse_categorical_crossentropy'
        if optimizer is None:
            optimizer = 'adam'
        if metrics is None:
            metrics = ['accuracy']

        self.tf_model.compile(loss=loss,
                              optimizer=optimizer,
                              metrics=metrics,
                              **kwargs)

    def fit(self, x_train, y_train, x_validate=None, y_validate=None,
            *,
            batch_size: int = 64,
            epochs: int = 5,
            callbacks=None,
            fit_kwargs={}):
        """
        Trains the model for a given number of epochs with given texts set list.
        """
        train_gen = DataGenerator(x_train, y_train)
        if x_validate is not None:
            valid_gen = DataGenerator(x_validate, y_validate)
        else:
            valid_gen = None
        return self.fit_generator(train_sample_gen=train_gen,
                                  valid_sample_gen=valid_gen,
                                  batch_size=batch_size,
                                  epochs=epochs,
                                  callbacks=callbacks,
                                  fit_kwargs=fit_kwargs)

    def fit_generator(self, train_sample_gen, valid_sample_gen=None,
                      *,
                      batch_size: int = 64,
                      epochs: int = 5,
                      callbacks=None,
                      fit_kwargs: Dict = {}):

        self.build_model_generator([g for g in [train_sample_gen, valid_sample_gen] if g])

        model_summary = []
        self.tf_model.summary(print_fn=lambda x: model_summary.append(x))
        logger.debug('\n'.join(model_summary))

        train_set = train_sample_gen
        train_set.label_indexer = self.label_indexer
        train_set.batch_size = batch_size

        if fit_kwargs is None:
            fit_kwargs = {}
        valid_gen = None
        if valid_sample_gen:
            valid_gen = train_sample_gen
            valid_gen.label_indexer = self.label_indexer
            valid_gen.batch_size = batch_size

            fit_kwargs['validation_data'] = valid_gen
            fit_kwargs['validation_steps'] = len(valid_gen)

        return self.tf_model.fit(train_set,
                                 epochs=epochs,
                                 callbacks=callbacks,
                                 **fit_kwargs)

    def predict(self, x_data, *,
                batch_size: int = 32,
                truncating: bool = False,
                multi_label_threshold: float = 0.5,
                predict_kwargs: Dict = None):
        """
        Generates output predictions for the input samples.

        Computation is done in batches.

        Args:
            x_data: The input texts, as a Numpy array (or list of Numpy arrays if the model has multiple inputs).
            batch_size: Integer. If unspecified, it will default to 32.
            truncating: remove values from sequences larger than `model.embedding.sequence_length`
            multi_label_threshold:
            predict_kwargs: arguments passed to ``predict()`` function of ``tf.keras.Model``

        Returns:
            array(s) of predictions.
        """
        if predict_kwargs is None:
            predict_kwargs = {}
        with kolibri.dnn.custom_object_scope():
            if truncating:
                seq_length = self.sequence_length
            else:
                seq_length = None

            # for i in range(len(self.wav_paths)):
            #     rate, wav = wavfile.read(self.wav_paths[i])
            #     yield wav.reshape(1, -1), self.y_values[i]
            tensor = x_data
            logger.debug(f'predict input shape {np.array(tensor).shape} x: \n{tensor}')
            pred = self.tf_model.predict(tensor, batch_size=batch_size, **predict_kwargs)
            logger.debug(f'predict output shape {pred.shape}')
            if self.multi_label:
                multi_label_binarizer = self.label_indexer.multi_label_binarizer  # type: ignore
                res = multi_label_binarizer.inverse_transform(pred,
                                                              threshold=multi_label_threshold)
            else:
                pred_argmax = pred.argmax(-1)
                lengths = [len(sen) for sen in x_data]
                res = self.label_indexer.inverse_transform(pred_argmax,
                                                           lengths=lengths)
                logger.debug(f'predict output argmax: {pred_argmax}')

        return res

    def evaluate(self, x_data, y_data, *,
                 batch_size: int = 32,
                 digits: int = 4,
                 multi_label_threshold: float = 0.5,
                 truncating: bool = False, ):
        y_pred = self.predict(x_data,
                              batch_size=batch_size,
                              truncating=truncating,
                              multi_label_threshold=multi_label_threshold)

        if self.multi_label:
            report = multi_label_classification_report(y_data,  # type: ignore
                                                       y_pred,  # type: ignore
                                                       binarizer=self.label_indexer.multi_label_binarizer)  # type: ignore

        else:
            original_report = sklearn_metrics.classification_report(y_data,
                                                                    y_pred,
                                                                    output_dict=True,
                                                                    digits=digits)
            print(sklearn_metrics.classification_report(y_data,
                                                        y_pred,
                                                        output_dict=False,
                                                        digits=digits))
            report = {
                'detail': original_report,
                **original_report['weighted avg']
            }
        return report

    @classmethod
    def load_model(cls, model_path):
        from kolibri.dnn.layers import Melspectrogram, Normalization2D
        model_config_path = os.path.join(model_path, 'model_config.json')
        model_config = json.loads(open(model_config_path, 'r').read())
        model = load_data_object(model_config)
        model.label_indexer = load_data_object(model_config['label_indexer'])

        tf_model_str = json.dumps(model_config['tf_model'])
        model.tf_model = tf.keras.models.model_from_json(tf_model_str, {'Melspectrogram': Melspectrogram,
                                                                        'Normalization2D': Normalization2D})

        model.tf_model.load_weights(os.path.join(model_path, 'model_weights.h5'))

        return model


if __name__ == "__main__":
    pass



import tensorflow as tf

from kolibri.dnn.embeddings.base_embedding import Embedding


class GRUDecoder(tf.keras.Model):
    def __init__(self,
                 embedding: Embedding,
                 hidden_size: int,
                 vocab_size: int):
        super(GRUDecoder, self).__init__()
        self.embedding = embedding

        self.gru = tf.keras.layers.GRU(hidden_size,
                                       return_sequences=True,
                                       return_state=True,
                                       recurrent_initializer='glorot_uniform')
        self.fc = tf.keras.layers.Dense(vocab_size)

    def call(self, dec_input, dec_hidden, enc_output):
        # The shape of x after passing through the embedding layer == (batch size, 1, embedding dimension)
        decoder_embedding = self.embedding.embed_model(dec_input)

        s = self.gru(decoder_embedding, initial_state=dec_hidden)
        decoder_outputs, decoder_state = s

        # Output shape == (batch size * 1, hidden layer size)
        output = tf.reshape(decoder_outputs, (-1, decoder_outputs.shape[2]))

        # Output shape == (batch size, vocab)
        x = self.fc(output)
        return x, decoder_state, None


if __name__ == "__main__":
    pass



import os
from glob import glob

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from scipy.io import wavfile


def envelope(y, rate, threshold):
    mask = []
    y = pd.Series(y).apply(np.abs)
    y_mean = y.rolling(window=int(rate / 20),
                       min_periods=1,
                       center=True).max()
    for mean in y_mean:
        if mean > float(threshold):
            mask.append(True)
        else:
            mask.append(False)
    return mask, y_mean


def save_sample(sample, rate, target_dir, fn, ix):
    fn = fn.split('.wav')[0]
    dst_path = os.path.join(target_dir.split('.')[0], fn + '_{}.wav'.format(str(ix)))
    if os.path.exists(dst_path):
        return
    wavfile.write(dst_path, rate, sample)


def split_wavs(wav, rate, delta_time, threshold):
    wav = downsample_mono(wav, rate)
    mask, y_mean = envelope(wav, rate, threshold=threshold)
    wav = wav[mask]
    delta_sample = int(delta_time * rate)
    samples = []
    # cleaned audio is less than a single sample
    # pad with zeros to delta_sample size
    if wav.shape[0] < delta_sample:
        sample = np.zeros(shape=(delta_sample,), dtype=np.int16)
        sample[:wav.shape[0]] = wav
        samples.append(sample)
    # step through audio and save every delta_sample
    # discard the ending audio if it is too short
    else:
        trunc = wav.shape[0] % delta_sample
        for cnt, i in enumerate(np.arange(0, wav.shape[0] - trunc, delta_sample)):
            start = int(i)
            stop = int(i + delta_sample)
            sample = wav[start:stop]
            samples.append(sample)
    return samples


def test_threshold(args):
    src_root = args.src_root
    wav_paths = glob('{}/**'.format(src_root), recursive=True)
    wav_path = [x for x in wav_paths if args.fn in x]
    if len(wav_path) != 1:
        print('audio file not found for sub-string: {}'.format(args.fn))
        return
    rate, wav = downsample_mono(wav_path[0], args.sr)
    mask, env = envelope(wav, rate, threshold=args.threshold)
    plt.style.use('ggplot')
    plt.title('Signal Envelope, Threshold = {}'.format(str(args.threshold)))
    plt.plot(wav[np.logical_not(mask)], color='r', label='remove')
    plt.plot(wav[mask], color='c', label='keep')
    plt.plot(env, color='m', label='envelope')
    plt.grid(False)
    plt.legend(loc='best')
    plt.show()



from tensorflow.keras import regularizers
from tensorflow.keras.layers import Input, Dense
from tensorflow.keras.models import Model

from kolibri.dnn.autoencoder.base_autoencoder import BaseAutoencoder


class DeepAutoencoder(BaseAutoencoder):

    @classmethod
    def get_default_hyper_parameters(cls):
        return {
            'nb-layers': 2,
            'sparsity': 0.01,
            'hidden': {
                'units': 100,
                'activation': 'relu'
            },
            'latent': {
                'units': 10,
                'activation': 'relu'
            },
            'output': {
                'activation': 'tanh'
            }
        }

    def build_model_arc(self):
        dim_in = self.input_dim


        #dim_out=self.hyper_parameters['latent']['units']
        dims_encoder = self.hyper_parameters['nb-layers']
        dims_decoding = dims_encoder
        input_img: object = Input(shape=(dim_in,), name='EncoderIn')

        encoded = input_img

        # Construct encoder layers
        out_dim = self.input_dim
        for i in range(dims_encoder):
            name = 'Encoder{0}'.format(i)
            out_dim = int(out_dim / 2)
            encoded = Dense(out_dim, activation='relu', name=name, activity_regularizer=regularizers.l2(self.hyper_parameters['sparsity']))(encoded)

        # Construct decoder layers
        # The decoded is connected to the encoders, whereas the decoder is not
        decoded = encoded
        decoder_input = Input(shape=(out_dim,), name='DecoderIn')

        decoder = decoder_input
        for i in range(dims_decoding):
            name = 'Decoder{0}'.format(i)

            activation = 'relu'
            out_dim = out_dim * 2
            decoder = Dense(out_dim, activation=activation, name=name, activity_regularizer=regularizers.l2(self.hyper_parameters['sparsity']))(decoder)

        decoder = Dense(self.input_dim, activation=activation, name='output')(decoder)

        self.encoder = Model(inputs=input_img, outputs=encoded)

        # instantiate decoder model

        self.decoder = Model(decoder_input, decoder, name='decoder')

        # autoencoder = encoder + decoder
        # instantiate autoencoder model
        self.autoencoder = Model(input_img,
                                 self.decoder(self.encoder(input_img)),
                                 name='autoencoder')

#		self.encoder = Model(input=input_img, output=encoded)
#		self.decoder = Model(input=decoder_input, output=decoder)

#		self.autoencoder.compile(optimizer='adadelta', loss='binary_crossentropy')



import copy

import numpy as np
from sklearn.base import ClassifierMixin
from sklearn.preprocessing import normalize

from kolibri import Component
from kolibri.utils import check_random_state
from kolibri.utils import get_dimensions


class HalfSpaceTrees(Component, ClassifierMixin):
    """Half--Space Trees.

    Implementation of the Streaming Half--Space--Trees (HS--Trees) [1]_, a fast one-class anomaly detector
    for evolving texts streams. It requires only normal texts for training and works well when anomalous
    texts are rare. The model features an ensemble of random HS--Trees, and the tree structure is
    constructed without any texts. This makes the method highly efficient because it requires no model
    restructuring when adapting to evolving texts streams.

    Parameters
    ----------

    n_features: int, required
        The dimensionality of the stream.

    n_estimators: int, optional (default=25)
       Number of trees in the ensemble.
       't' in the original paper.

    window_size: int, optional (default=250)
        The window size of the stream.
        'Psi' in the original paper.

    depth: int, optional (default=15)
        The maximum depth of the trees in the ensemble.
        'maxDepth' in the original paper.

    size_limit: int, optional (default=50)
        The minimum mass required in a node (as a fraction of the window size) to calculate the anomaly score.
        'sizeLimit' in the original paper.
        A good setting is 0.1 * window_size

    anomaly_threshold: double, optional (default=0.5)
        The threshold for declaring anomalies.
        Any instance prediction probability above this threshold will be declared as an anomaly.

    random_state: int, RandomState instance or None, optional (default=None)
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used
        by `np.random`.

    References
    ----------
    .. [1] S.C.Tan, K.M.Ting, and T.F.Liu, “Fast anomaly detection for streaming texts,”
       in IJCAI Proceedings - International Joint Conference on Artificial Intelligence,
       2011, vol. 22, no. 1, pp. 1511–1516.

    Examples
    --------
    .. code-block:: python

       # Imports
       from skmultiflow.texts import SEAGenerator
       from skmultiflow.anomaly_detection import HalfSpaceTrees

       # Setup a texts stream
       stream = SEAGenerator(random_state=1)

       # Setup Half-Space Trees estimator
       half_space_trees = HalfSpaceTrees()

       # Setup variables to control loop and track performance
       n_samples = 0
       correct_cnt = 0

       # Train the estimator(s) with the samples provided by the texts stream
       while n_samples < max_samples and stream.has_more_samples():
           X, y = stream.next_sample()
           y_pred = half_space_trees.predict(X) # Error is thrown here
           if y[0] == y_pred[0]:
               correct_cnt += 1
           half_space_trees = half_space_trees.partial_fit(X, y)
           n_samples += 1

       # Display results
       print('{} samples analyzed.'.format(n_samples))
       print('Half-Space Trees accuracy: {}'.format(correct_cnt / n_samples))
    """

    def __init__(self,
                 n_features,
                 window_size=250,
                 depth=15,
                 n_estimators=25,
                 size_limit=50,
                 anomaly_threshold=0.5,
                 random_state=None):

        super().__init__()
        self.window_size = window_size
        self.depth = depth
        self.n_estimators = n_estimators
        self.min_values = []
        self.max_values = []
        self.ensemble = []
        self.n_features = n_features
        self.size_limit = size_limit
        self.samples_seen = 0
        self.anomaly_threshold = anomaly_threshold
        self.is_learning_phase_on = True
        self.random_state = random_state
        self._random_state = None

    def partial_fit(self, X, y, classes=None, sample_weight=None):
        """ Partially (incrementally) fit the model.

        Parameters
        ----------
        X : numpy.ndarray of shape (n_samples, n_features)
            The features to train the model.

        y: numpy.ndarray of shape (n_samples)
            An array-like with the class y_values of all samples in X.

        classes: None
            Not used by this method.

        sample_weight: None
            Not used by this method.

        Returns
        -------
        self

        """

        row_cnt, _ = X.shape

        if self.samples_seen == 0:
            self._random_state = check_random_state(self.random_state)
            self.build_trees()

        for i in range(row_cnt):
            self._partial_fit(X[i], y[i])

        return self

    def _partial_fit(self, X, y):
        """ Trains the model on samples X and corresponding targets y.

        Private function where actual training is carried on.

        Parameters
        ----------
        X: numpy.ndarray of shape (1, n_features)
            Instance attributes.

        y: int
            Class label for sample X.

        """

        # Populates first reference mass profile for every tree defined in the ensemble.
        if self.samples_seen < self.window_size:
            self.update_mass(X, True)
        # Otherwise populates latest mass profile for every tree defined in the ensemble.
        else:
            self.update_mass(X, False)

        if self.is_learning_phase_on is True and self.samples_seen > self.window_size:
            self.set_is_learning_phase_on(False)

        if (self.samples_seen % self.window_size) == 0:
            self.update_models()

        self.samples_seen += 1

    def predict(self, X):
        """ Predict classes for the passed texts.

        Parameters
        ----------
        X : numpy.ndarray of shape (n_samples, n_features)
            The set of texts samples to predict the class y_values for.

        Returns
        -------
        A numpy.ndarray with all the predictions for the samples in X.

        """
        r, _ = get_dimensions(X)
        predictions = []
        for i in range(r):
            y_proba = self.predict_proba(X)
            if y_proba is None:
                # Ensemble is empty, all classes equal, default to zero
                predictions.append(0)
            else:
                # if prediction of this instance being anomaly is greater than the threshold defined,
                # then this instance is classified as an anomaly.
                if y_proba[0][1] > self.anomaly_threshold:
                    predictions.append(1)
                else:
                    predictions.append(0)
        return np.asarray(predictions)

    def predict_proba(self, X):
        """ Estimates the probability of each sample in X belonging to each of the class-y_values (normal and outlier).

        Class probabilities are calculated as the mean predicted class probabilities per base estimator.

        Parameters
        ----------
         X: numpy.ndarray of shape (n_samples, n_features)
            Samples for which we want to predict the class probabilities.

        Returns
        -------
        numpy.ndarray of shape (n_samples, n_classes)
            Predicted class probabilities for all instances in X.
            Class probabilities for a sample shall sum to 1 as long as at least one estimators has non-zero predictions.
            If no estimator can predict probabilities, probabilities of 0 are returned.
        """
        y_proba_mean = None
        max_score = self.window_size * pow(2.0, self.depth)

        for i in range(self.n_estimators):
            y_proba = self.ensemble[i].predict_proba(X, max_score)
            if y_proba_mean is None:
                y_proba_mean = y_proba
            else:
                y_proba_mean = y_proba_mean + (y_proba - y_proba_mean) / (i + 1)
        return normalize(y_proba_mean, norm='l1')

    def initialise_work_space(self):
        """
        Initialises work spaces.
        For every dimension in the feature space, creates a minimum and a maximum work range.
        """
        for i in range(self.n_features):
            sq = self._random_state.uniform(0, 1)
            min_element = sq - 2 * max(sq, 1 - sq)
            max_element = sq + 2 * max(sq, 1 - sq)
            try:
                self.min_values[i] = min_element
            except IndexError:
                self.min_values.append(min_element)
            try:
                self.max_values[i] = max_element
            except IndexError:
                self.max_values.append(max_element)

    def build_trees(self):
        """
        Initialises ensemble.
        """
        for i in range(self.n_estimators):
            self.initialise_work_space()
            tree = HalfSpaceTree(self.depth, self.n_features, self.size_limit, self.min_values, self.max_values,
                                 self._random_state)
            self.ensemble.append(tree)

    def update_mass(self, X, boolean):
        """
        Populates mass profiles for every tree defined in the ensemble.

        Parameters
        ----------
        X: numpy.ndarray of shape (1, n_features)
            Instance attributes.

        boolean: boolean, True or False
            True to update reference mass, False to update latest mass
        """
        for i in range(self.n_estimators):
            self.ensemble[i].update_mass(X, boolean)

    def update_models(self):
        """
        Updates the mass profile of every tree in the ensemble.

        """
        for i in range(self.n_estimators):
            tree = self.ensemble[i]
            tree.update_model(tree.root)

    def set_is_learning_phase_on(self, boolean):
        """
        Sets learning phase in each tree defined in the ensemble.

        Parameters
        ----------
        boolean: Boolean

        """
        self.is_learning_phase_on = boolean
        for i in range(self.n_estimators):
            self.ensemble[i].is_learning_phase_on = boolean


class HalfSpaceTree:

    def __init__(self, max_depth, n_features, size_limit, min_values, max_values, random_state=None):
        """
        Half Space Tree

        Parameters
        ----------
        n_features: int, required
            The dimensionality of the stream.

        max_depth: int, optional (default=15)
            The maximum depth of the trees in the ensemble.
            'maxDepth' in the original paper.

        size_limit: int, optional (default=50)
            The minimum mass required in a node (as a fraction of the window size) to calculate the anomaly score.
            'sizeLimit' in the original paper.

        min_values: Array of floats
            Minimum work range for every dimension.

        max_values: Array of floats
            Maximum work range for every dimension.

        random_state: int, RandomState instance or None, optional (default=None)
            If int, random_state is the seed used by the random number generator;
            If RandomState instance, random_state is the random number generator;
            If None, the random number generator is the RandomState instance used
            by `np.random`.
        """
        super().__init__()
        self.max_depth = max_depth
        self.n_features = n_features
        self.size_limit = size_limit
        self.is_learning_phase_on = True
        self.random_state = random_state
        self._random_state = check_random_state(random_state)
        self.root = self.build_tree(min_values, max_values)

    def predict_proba(self, X, max_score):
        """ Predicts probabilities of all label of the X instance(s)

        Parameters
        ----------
        X: numpy.ndarray of shape (n_samples, n_features)
            Samples for which we want to predict the y_values.

        max_score: float
            The maximum score of an instance could have in the tree.

        Returns
        -------
        numpy.array
            Predicted the probabilities of all the y_values for all instances in X.

        """
        r, _ = get_dimensions(X)
        predictions = []
        for i in range(r):
            votes = copy.deepcopy(self.get_votes_for_instance(X[i], max_score))
            y_proba = np.zeros(int(max(votes.keys())) + 1)
            for key, value in votes.items():
                y_proba[int(key)] = value
            predictions.append(y_proba)
        return np.array(predictions)

    def get_votes_for_instance(self, X, max_score):
        """ Get class votes for a single instance.

        Parameters
        ----------
        X: numpy.ndarray of length_train equal to the number of features.
            Instance attributes.

        max_score: float
            The maximum score of an instance could have in the tree.
        Returns
        -------
        dict (class_value, weight)

        """
        if self.root is not None and self.is_learning_phase_on is not True:
            score = self.anomaly_score(X)
            anomaly_score = 1 - score / max_score
            return {0: 1 - anomaly_score, 1: anomaly_score}
        else:
            return {0: 0.5, 1: 0.5}

    def anomaly_score(self, X):
        """
        Calculates anomaly score of a given instance X.

        Parameters
        ----------
        X: numpy.ndarray of shape (1, n_features)
            Instance attributes.

        Returns
        -------

        """
        return self._anomaly_score(self.root, X)

    def _anomaly_score(self, node, X):
        """
        Private function to calculate anomaly score.

        Parameters
        ----------
        node: HalfSpaceTreeNode
            Current node

        X: numpy.ndarray of shape (1, n_features)
            Instance attributes.

        Returns
        -------
        anomaly_score: float
            Anomaly score of the instance X.

        """
        if node.internal_node is not True or node.r <= self.size_limit:
            return node.r * pow(2.0, node.depth)
        else:
            if X[node.split_attribute] > node.split_value:
                return self._anomaly_score(node.right, X)
            else:
                return self._anomaly_score(node.left, X)

    def build_tree(self, min_values, max_values, current_depth=1):
        """
        Builds a single half space tree.

        Parameters
        ----------
        min_values: Array of floats
            Minimum work range for every dimension.
        max_values: Array of floats
            Maximum work range for every dimension.
        current_depth:
            Indicates the current depth of the tree.
        Returns
        -------
        HalfSpaceTreeNode:
            root of the tree.
        """
        if self.max_depth == current_depth:
            return HalfSpaceTreeNode(depth=current_depth, internal_node=False)
        else:
            random_feature_idx = self._random_state.randint(0, self.n_features - 1)
            p = (min_values[random_feature_idx] + max_values[random_feature_idx]) / 2.0
            temp = max_values[random_feature_idx]
            max_values[random_feature_idx] = p
            left = self.build_tree(min_values, max_values, current_depth + 1)
            max_values[random_feature_idx] = temp
            min_values[random_feature_idx] = p
            right = self.build_tree(min_values, max_values, current_depth + 1)
            return HalfSpaceTreeNode(left=left, right=right, split_attribute=random_feature_idx,
                                     split_value=p, depth=current_depth)

    def update_mass(self, X, is_reference_window):
        """
        Records mass texts in the tree.

        Parameters
        ----------
        X: numpy.ndarray of shape (1, n_features)
            Instance attributes.
        is_reference_window: Boolean
            True if the classifier is in the first reference window, otherwise false
        Returns
        -------

        """
        return self._update_mass(X, self.root, is_reference_window)

    def _update_mass(self, X, node, is_reference_window):
        """
        Private function where actual mass texts is recorded.

        Parameters
        ----------
        X: numpy.ndarray of shape (1, n_features)
            Instance attributes.
        node: HalfSpaceTreeNode

        is_reference_window: Boolean
            True if the classifier is in the first reference window, otherwise false

        """
        if node is None:
            return

        if is_reference_window:
            node.r = node.r + 1
        else:
            node.l = node.l + 1

        if node.depth < self.max_depth:
            if X[node.split_attribute] > node.split_value:
                self._update_mass(X, node.right, is_reference_window)
            else:
                self._update_mass(X, node.left, is_reference_window)
        else:
            return

    def update_model(self, node):
        """
        Updates model to adapt to new texts.

        Parameters
        ----------
        node: HalfSpaceTreeNode

        """
        if node is not None:
            if node.r is not 0 or node.l is not 0:
                node.r = node.l

            if node.l is not 0:
                node.l = 0

            self.update_model(node.left)
            self.update_model(node.right)
        return


class HalfSpaceTreeNode:

    def __init__(self, left=None, right=None, split_value=0.0, split_attribute=-1,
                 depth=1, internal_node=True):
        """
        Parameters
        ----------
        left: HalfSpaceTreeNode, optional (default=None)
            The left node pointed to this node.

        right: HalfSpaceTreeNode, optional (default=None)
            The left node pointed to this node.

        split_value: double, optional (default=0.0)
            Mid point of the selected random dimension.
            'p' in the original paper.

        split_attribute: int, optional (default=-1)
            Randomly selected attribute for the node.
            'q' in the original paper.

        depth: int, optional (default=1)
            Represents depth of the node in the tree.
            'k' in the original paper.

        internal_node: boolean, optional (default=True)
            True if node is an internal node, false if it is an leaf node.

        Attributes
        ----------
        l: int
            Mass of the node in the reference window.

        r: int
            Mass of the node in the latest window.

        """
        self.left = left
        self.right = right
        self.l = 0
        self.r = 0
        self.depth = depth
        self.split_value = split_value
        self.split_attribute = split_attribute
        self.internal_node = internal_node



import logging

from kolibri import utils
from kolibri.application.search import *

logger = logging.getLogger(__name__)


class App(object):
    """
    A conversational dialog chat bot.
    """

    def __init__(self, name, **kwargs):
        self.name = name

        database_adapter = kwargs.get('database_adapter', 'kolibri.application.datastore.SQLDatabaseAdapter')

        self.datastore = utils.misc.initialize_class(database_adapter, **kwargs)

        postprocessors = kwargs.get('postprocessor', [
            'kolibri.application.score.BestMatch'
        ])

        primary_search_algorithm = IndexedTextSearch(self, **kwargs)
        text_search_algorithm = TextSearch(self, **kwargs)

        self.search_algorithms = {
            primary_search_algorithm.name: primary_search_algorithm,
            text_search_algorithm.name: text_search_algorithm
        }
        self.postprocessors = []

        for postprocessor in postprocessors:
            postprocessor_class = utils.initialize_class(postprocessor, self, **kwargs)
            self.postprocessors.append(postprocessor_class)

        self.logger = logger
        # Allow the app to save input it receives so that it can learn
        self.read_only = kwargs.get('read_only', False)
